import pytest
import torch

from fme.core.atmosphere_data import (
    AtmosphereData,
    _height_at_interface,
    compute_layer_thickness,
)
from fme.core.constants import LATENT_HEAT_OF_FREEZING


def test_compute_layer_thickness():
    pressure_at_interface = torch.tensor(
        [
            [[1, 2, 3], [4, 5, 6]],
            [[1, 2, 3], [4, 5, 6]],
        ]
    )
    air_temperature = torch.tensor(
        [
            [[300, 310], [300, 310]],
            [[300, 310], [300, 310]],
        ]
    )
    specific_total_water = torch.full((2, 2, 2), 0.1)
    dz = compute_layer_thickness(
        pressure_at_interface, air_temperature, specific_total_water
    )
    assert dz.shape == (2, 2, 2)
    assert torch.all(dz >= 0.0)


def test__height_at_interface():
    layer_thickness = torch.tensor([[[3, 2], [1, 0.5]], [[3, 2], [1, 0.5]]])
    height_at_surface = torch.tensor([[10, 20], [10, 20]])
    height_at_interface = _height_at_interface(layer_thickness, height_at_surface)
    assert height_at_interface.shape == (2, 2, 3)
    assert torch.equal(
        height_at_interface,
        torch.tensor(
            [[[15, 12, 10], [21.5, 20.5, 20]], [[15, 12, 10], [21.5, 20.5, 20]]]
        ),
    )


@pytest.mark.parametrize("has_water_variable", [True, False])
def test_missing_specific_total_water(has_water_variable):
    """Check shape of specific total water and make sure that it returns None
    when it can't be computed."""
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8

    def _get_data(water: bool):
        if water:
            return {
                "x": torch.rand(n_samples, n_time_steps, nlat, nlon),
                "specific_total_water_0": torch.rand(
                    n_samples, n_time_steps, nlat, nlon
                ),
                "specific_total_water_1": torch.rand(
                    n_samples, n_time_steps, nlat, nlon
                ),
                "PRESsfc": torch.rand(n_samples, n_time_steps, nlat, nlon),
            }
        else:
            return {
                "x": torch.rand(n_samples, n_time_steps, nlat, nlon),
                "PRESsfc": torch.rand(n_samples, n_time_steps, nlat, nlon),
            }

    atmos_data = AtmosphereData(_get_data(water=has_water_variable))

    if has_water_variable:
        assert atmos_data.specific_total_water is not None
        assert atmos_data.specific_total_water.shape == (
            n_samples,
            n_time_steps,
            nlat,
            nlon,
            2,
        )
    else:
        with pytest.raises(KeyError):
            _ = atmos_data.specific_total_water


@pytest.mark.parametrize("missing_water_layer", [True, False])
def test_keyerror_when_missing_specific_total_water_layer(missing_water_layer: bool):
    """Check shape of specific total water and make sure that it returns None
    when it can't be computed."""
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8

    def _get_data(missing_water_layer: bool):
        data = {
            "x": torch.rand(n_samples, n_time_steps, nlat, nlon),
            "specific_total_water_1": torch.rand(n_samples, n_time_steps, nlat, nlon),
            "PRESsfc": torch.rand(n_samples, n_time_steps, nlat, nlon),
        }
        if not missing_water_layer:
            data["specific_total_water_0"] = torch.rand(
                n_samples, n_time_steps, nlat, nlon
            )
        return data

    atmos_data = AtmosphereData(_get_data(missing_water_layer))

    if not missing_water_layer:
        assert atmos_data.specific_total_water is not None
        assert atmos_data.specific_total_water.shape == (
            n_samples,
            n_time_steps,
            nlat,
            nlon,
            2,
        )
    else:
        with pytest.raises(ValueError):
            _ = atmos_data.specific_total_water


@pytest.mark.parametrize("has_frozen_precipitation", [False, True])
def test_net_surface_energy_flux(has_frozen_precipitation):
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8
    shape = (n_samples, n_time_steps, nlat, nlon)

    def _get_data(
        shape: tuple[int, ...], has_frozen_precipitation: bool
    ) -> AtmosphereData:
        ones = torch.ones(shape)
        surface_pressure = {"PRESsfc": ones}
        energy_fluxes = {
            "LHTFLsfc": ones,
            "SHTFLsfc": ones,
            "ULWRFsfc": ones,
            "DLWRFsfc": ones,
            "USWRFsfc": ones,
            "DSWRFsfc": ones,
        }
        if has_frozen_precipitation:
            frozen_precipitation = {
                "GRAUPELsfc": ones / LATENT_HEAT_OF_FREEZING,
                "ICEsfc": ones / LATENT_HEAT_OF_FREEZING,
                "SNOWsfc": ones / LATENT_HEAT_OF_FREEZING,
            }
        else:
            frozen_precipitation = {}
        data = surface_pressure | energy_fluxes | frozen_precipitation
        return AtmosphereData(data)

    atmosphere_data = _get_data(shape, has_frozen_precipitation)

    if has_frozen_precipitation:
        expected = torch.full(shape, -5)
    else:
        expected = torch.full(shape, -2)

    result = atmosphere_data.net_surface_energy_flux
    torch.testing.assert_close(result, expected, check_dtype=False)


def test_net_energy_flux_into_atmosphere():
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8
    shape = (n_samples, n_time_steps, nlat, nlon)

    def _get_data(shape: tuple[int, ...]) -> AtmosphereData:
        ones = torch.ones(shape, dtype=torch.float32)
        surface_pressure = {"PRESsfc": ones}
        energy_fluxes = {
            "LHTFLsfc": ones,
            "SHTFLsfc": ones,
            "ULWRFsfc": ones,
            "DLWRFsfc": ones,
            "USWRFsfc": ones,
            "DSWRFsfc": ones,
            "ULWRFtoa": ones,
            "USWRFtoa": ones,
            "DSWRFtoa": ones,
        }
        data = surface_pressure | energy_fluxes
        return AtmosphereData(data)

    atmosphere_data = _get_data(shape)

    expected = torch.full(shape, 1.0, dtype=torch.float32)

    result = atmosphere_data.net_energy_flux_into_atmosphere
    torch.testing.assert_close(result, expected)


@pytest.mark.parametrize(
    ("fields", "expected_sum"),
    [
        (("GRAUPELsfc", "ICEsfc", "SNOWsfc"), 3),
        (("GRAUPELsfc", "ICEsfc"), 0),
        (("total_frozen_precipitation_rate",), 1),
        ((), 0),
    ],
    ids=lambda x: f"{x!r}",
)
def test_frozen_precipitation_rate(fields, expected_sum):
    n_samples, n_time_steps, nlat, nlon = 2, 3, 4, 8
    shape = (n_samples, n_time_steps, nlat, nlon)

    def _get_data(shape: tuple[int, ...]) -> AtmosphereData:
        ones = torch.ones(shape)
        surface_pressure = {"PRESsfc": ones}
        frozen_precipitation = {}
        for field in fields:
            frozen_precipitation[field] = ones
        data = surface_pressure | frozen_precipitation
        return AtmosphereData(data)

    atmosphere_data = _get_data(shape)
    expected = torch.full(shape, expected_sum)

    result = atmosphere_data.frozen_precipitation_rate
    torch.testing.assert_close(result, expected, check_dtype=False)


def test_windspeed_at_10m():
    shape = (2, 4)
    data_dict = {"UGRD10m": torch.ones(shape), "VGRD10m": -2 * torch.ones(shape)}
    atmosphere_data = AtmosphereData(data_dict)
    expected = torch.sqrt(torch.full(shape, 5))
    result = atmosphere_data.windspeed_at_10m
    torch.testing.assert_close(result, expected)



import dataclasses
import datetime
import logging
import os
import pathlib
import tempfile
from collections.abc import Iterable
from unittest.mock import patch

import dacite
import numpy as np
import pytest
import torch
import xarray as xr
import yaml

from fme.ace.aggregator.inference import InferenceEvaluatorAggregatorConfig
from fme.ace.data_loading.inference import (
    InferenceDataLoaderConfig,
    InferenceInitialConditionIndices,
)
from fme.ace.inference.data_writer import DataWriterConfig
from fme.ace.inference.data_writer.time_coarsen import TimeCoarsenConfig
from fme.ace.inference.evaluator import (
    InferenceEvaluatorConfig,
    StepperOverrideConfig,
    main,
    resolve_variable_metadata,
)
from fme.ace.registry import ModuleSelector
from fme.ace.stepper import SingleModuleStepperConfig, Stepper, TrainOutput
from fme.ace.stepper.single_module import StepperConfig
from fme.ace.testing import DimSizes, FV3GFSData, MonthlyReferenceData
from fme.core import metrics
from fme.core.coordinates import (
    DimSize,
    HybridSigmaPressureCoordinate,
    LatLonCoordinates,
)
from fme.core.dataset.config import XarrayDataConfig
from fme.core.dataset.data_typing import VariableMetadata
from fme.core.dataset_info import DatasetInfo
from fme.core.derived_variables import compute_derived_quantities
from fme.core.device import get_device
from fme.core.logging_utils import LoggingConfig
from fme.core.multi_call import MultiCallConfig
from fme.core.normalizer import NormalizationConfig
from fme.core.ocean import Ocean, OceanConfig
from fme.core.step.multi_call import MultiCallStep, MultiCallStepConfig
from fme.core.step.single_module import SingleModuleStep, SingleModuleStepConfig
from fme.core.testing import mock_wandb
from fme.core.typing_ import EnsembleTensorDict, TensorDict, TensorMapping

DIR = pathlib.Path(__file__).parent
TIMESTEP = datetime.timedelta(hours=6)


class PlusOne(torch.nn.Module):
    def forward(self, x):
        return x + 1


def save_plus_one_stepper(
    path: pathlib.Path,
    in_names: list[str],
    out_names: list[str],
    mean: float,
    std: float,
    data_shape: list[int],
    normalization_names: Iterable[str] | None = None,
    timestep: datetime.timedelta = TIMESTEP,
    nz_interface: int = 7,
    ocean=None,
    multi_call: MultiCallConfig | None = None,
):
    if multi_call is None:
        all_names = list(set(in_names).union(out_names))
    else:
        all_names = list(set(in_names).union(out_names)) + multi_call.names
    if normalization_names is None:
        normalization_names = all_names
    with tempfile.TemporaryDirectory() as temp_dir:
        mean_filename = pathlib.Path(temp_dir) / "means.nc"
        std_filename = pathlib.Path(temp_dir) / "stds.nc"
        xr.Dataset(
            {
                name: xr.DataArray(
                    mean,
                )
                for name in normalization_names
            }
        ).to_netcdf(mean_filename)
        xr.Dataset({name: xr.DataArray(std) for name in normalization_names}).to_netcdf(
            std_filename
        )
        config = SingleModuleStepperConfig(
            builder=ModuleSelector(type="prebuilt", config={"module": PlusOne()}),
            in_names=in_names,
            out_names=out_names,
            normalization=NormalizationConfig(
                global_means_path=str(mean_filename),
                global_stds_path=str(std_filename),
            ),
            ocean=ocean,
            multi_call=multi_call,
        )
        horizontal_coordinate = LatLonCoordinates(
            lat=torch.zeros(data_shape[-2]), lon=torch.zeros(data_shape[-1])
        )
        vertical_coordinate = HybridSigmaPressureCoordinate(
            ak=torch.arange(nz_interface), bk=torch.arange(nz_interface)
        )
        variable_metadata = {
            out_names[0]: VariableMetadata(
                units="cm",
                long_name="an output variable",
            ),
        }  # attach metadata to this output var to validate that persists in stepper
        dataset_info = DatasetInfo(
            horizontal_coordinates=horizontal_coordinate,
            vertical_coordinate=vertical_coordinate,
            timestep=timestep,
            variable_metadata=variable_metadata,
        )
        stepper = config.get_stepper(
            dataset_info=dataset_info,
        )
        torch.save({"stepper": stepper.get_state()}, path)
        mean_filename.unlink()
        std_filename.unlink()


def validate_stepper_ocean(stepper: Stepper, expected_ocean_config: OceanConfig | None):
    assert isinstance(stepper._step_obj, MultiCallStep)
    assert isinstance(stepper._step_obj._wrapped_step, SingleModuleStep)
    assert stepper._step_obj._wrapped_step._config.ocean == expected_ocean_config
    if expected_ocean_config is not None:
        assert isinstance(stepper._step_obj._wrapped_step.ocean, Ocean)
        assert (
            stepper._step_obj._wrapped_step.ocean.surface_temperature_name
            == expected_ocean_config.surface_temperature_name
        )
        assert (
            stepper._step_obj._wrapped_step.ocean.ocean_fraction_name
            == expected_ocean_config.ocean_fraction_name
        )
    else:
        assert stepper._step_obj._wrapped_step.ocean is None


def validate_stepper_multi_call(
    stepper: Stepper, expected_multi_call_config: MultiCallConfig | None
):
    assert isinstance(stepper._step_obj, MultiCallStep)
    assert isinstance(stepper._step_obj._wrapped_step, SingleModuleStep)
    assert stepper._step_obj._config.config == expected_multi_call_config


def test_inference_backwards_compatibility(tmp_path: pathlib.Path):
    """
    Inference test using a serialized model from an earlier commit, to ensure
    earlier models can be used with the updated inference code.
    """
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = DIR / "stepper_test_data"

    horizontal = [DimSize("lat", 4), DimSize("lon", 8)]
    dim_sizes = DimSizes(
        n_time=8,
        horizontal=horizontal,
        nz_interface=2,
    )
    if not stepper_path.exists():
        # this helps us to re-generate data if the stepper changes
        # to re-generate, just delete the data and run the test (it will fail)
        save_plus_one_stepper(
            stepper_path,
            in_names,
            out_names,
            mean=0.0,
            std=1.0,
            data_shape=dim_sizes.shape_nd,
        )
        assert False, "stepper_test_data did not exist, it has been created"
    use_prediction_data = False
    n_forward_steps = 2
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        timestep=TIMESTEP,
    )


@pytest.mark.parametrize(
    "use_prediction_data, n_forward_steps",
    [(True, 2), (False, int(30 / 20 * 36)), (False, 1)],
)
def test_inference_plus_one_model(
    tmp_path: pathlib.Path, use_prediction_data: bool, n_forward_steps: int
):
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("lat", 16), DimSize("lon", 32)]
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    if use_prediction_data:
        # use std of 10 so the stepper would have errors at the plus-one problem
        std = 10.0
    else:
        std = 1.0
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=std,
        data_shape=dim_sizes.shape_nd,
        timestep=datetime.timedelta(days=20),
    )
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        save_monthly_files=False,  # requires timestep == 6h
        timestep=datetime.timedelta(days=20),
    )


def inference_helper(
    tmp_path,
    in_names,
    out_names,
    use_prediction_data,
    dim_sizes: DimSizes,
    n_forward_steps,
    stepper_path,
    timestep: datetime.timedelta,
    save_monthly_files: bool = True,
    derived_names: list[str] = [],
    allow_incompatible_dataset_info: bool = True,  # stepper checkpoint has arbitrary info  # noqa: E501
):
    time_varying_values = [float(i) for i in range(dim_sizes.n_time)]
    all_names = list(set(in_names).union(out_names))
    forcing_names = list(set(in_names).difference(out_names))
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        time_varying_values=time_varying_values,
        timestep_days=timestep.total_seconds() / 86400,
        save_vertical_coordinate=False,
    )
    if use_prediction_data:
        prediction_data = data.inference_data_loader_config
    else:
        prediction_data = None

    if dim_sizes.n_time > 365 * 4:
        monthly_reference_filename = str(
            MonthlyReferenceData(
                path=pathlib.Path(tmp_path),
                names=all_names,
                dim_sizes=DimSizes(
                    n_time=48,
                    horizontal=dim_sizes.horizontal,
                    nz_interface=1,
                ),
                n_ensemble=3,
            ).data_filename
        )
    else:
        monthly_reference_filename = None
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        loader=data.inference_data_loader_config,
        prediction_loader=prediction_data,
        aggregator=InferenceEvaluatorAggregatorConfig(
            monthly_reference_data=monthly_reference_filename, log_video=True
        ),
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            log_extended_video_netcdfs=True,
            save_histogram_files=True,
            save_monthly_files=save_monthly_files,
        ),
        forward_steps_in_memory=1,
        allow_incompatible_dataset=allow_incompatible_dataset_info,
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(
            yaml_config=str(config_filename),
        )
        wandb_logs = wandb.get_logs()

    all_out_names = out_names + derived_names

    n_ic_timesteps = 1
    summary_log_step = 1
    assert len(wandb_logs) == n_ic_timesteps + config.n_forward_steps + summary_log_step
    for i in range(n_ic_timesteps + config.n_forward_steps):
        log = wandb_logs[i]
        for var in all_out_names:
            if i == 0 and var not in in_names:
                assert f"inference/mean/weighted_rmse/{var}" not in log
            else:
                # if these are off by something like 90% then probably the stepper
                # is being used instead of the prediction_data
                assert log[f"inference/mean/weighted_rmse/{var}"] == 0.0
                assert log[f"inference/mean/weighted_bias/{var}"] == 0.0

    example_output_var = out_names[0]
    # for all data written out, this example variable should have the same metadata
    # as the stepper, not the dataset on disk

    if not use_prediction_data:
        initial_condition_ds = xr.open_dataset(
            tmp_path / "initial_condition.nc", decode_timedelta=False
        )
        for dim_name in ["lat", "lon"]:
            assert dim_name in initial_condition_ds.dims
            assert dim_name in initial_condition_ds.data_vars[example_output_var].dims
            assert dim_name in initial_condition_ds.coords

    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc",
        decode_timedelta=False,
        decode_times=False,
    )
    assert len(prediction_ds["time"]) == config.n_forward_steps
    for i in range(config.n_forward_steps - 1):
        np.testing.assert_allclose(
            prediction_ds[example_output_var].isel(time=i).values + 1,
            prediction_ds[example_output_var].isel(time=i + 1).values,
        )
        assert not np.any(
            np.isnan(prediction_ds[example_output_var].isel(time=i + 1).values)
        )

    assert "lat" in prediction_ds.coords
    assert "lon" in prediction_ds.coords
    assert prediction_ds[out_names[0]].attrs == {
        "units": "cm",
        "long_name": "an output variable",
    }

    if use_prediction_data:
        assert not os.path.exists(tmp_path / "restart.nc")
        assert not os.path.exists(tmp_path / "initial_condition.nc")
    else:
        restart_ds = xr.open_dataset(
            tmp_path / "restart.nc", decode_timedelta=False, decode_times=False
        )
        np.testing.assert_allclose(
            prediction_ds[example_output_var].isel(time=-1).values,
            restart_ds[example_output_var].values,
        )

        ic_ds = xr.open_dataset(
            tmp_path / "initial_condition.nc",
            decode_timedelta=False,
            decode_times=False,
        )
        np.testing.assert_allclose(ic_ds[example_output_var].values, 0.0)

    metric_ds = xr.open_dataset(
        tmp_path / "reduced_autoregressive_predictions.nc", decode_timedelta=False
    )
    assert example_output_var in metric_ds.data_vars
    assert metric_ds.data_vars[example_output_var].attrs["units"] == "cm"
    assert (
        metric_ds.data_vars[example_output_var].attrs["long_name"]
        == f"ensemble mean of an output variable"
    )
    assert f"rmse_{example_output_var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"rmse_{example_output_var}"].attrs["units"] == "cm"
    assert (
        metric_ds.data_vars[f"rmse_{example_output_var}"].attrs["long_name"]
        == f"root mean squared error of an output variable"
    )
    assert f"bias_{example_output_var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"bias_{example_output_var}"].attrs["units"] == "cm"
    assert f"min_err_{example_output_var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"min_err_{example_output_var}"].attrs["units"] == "cm"
    assert f"max_err_{example_output_var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"max_err_{example_output_var}"].attrs["units"] == "cm"
    assert f"gen_var_{example_output_var}" in metric_ds.data_vars
    assert metric_ds.data_vars[f"gen_var_{example_output_var}"].attrs["units"] == ""
    assert (
        metric_ds.data_vars[f"gen_var_{example_output_var}"].attrs["long_name"]
        == f"prediction variance of an output variable as fraction of target variance"
    )
    assert "lat" in metric_ds.coords
    assert "lon" in metric_ds.coords

    time_mean_diagnostics = xr.open_dataset(
        tmp_path / "time_mean_diagnostics.nc", decode_timedelta=False
    )
    actual_var_names = sorted([str(k) for k in time_mean_diagnostics.keys()])
    assert len(actual_var_names) == 2 * len(all_out_names)
    assert f"bias_map-{example_output_var}" in actual_var_names
    assert (
        time_mean_diagnostics.data_vars[f"bias_map-{example_output_var}"].attrs["units"]
        == "cm"
    )
    assert f"gen_map-{example_output_var}" in actual_var_names
    assert (
        time_mean_diagnostics.data_vars[f"gen_map-{example_output_var}"].attrs["units"]
        == "cm"
    )
    assert len(time_mean_diagnostics.coords) == 2
    assert "lat" in time_mean_diagnostics.coords
    assert "lon" in time_mean_diagnostics.coords

    zonal_mean_diagnostics = xr.open_dataset(
        tmp_path / "zonal_mean_diagnostics.nc", decode_timedelta=False
    )
    actual_var_names = sorted([str(k) for k in zonal_mean_diagnostics.keys()])
    assert len(actual_var_names) == 2 * len(all_out_names)
    assert f"error-{example_output_var}" in actual_var_names
    assert (
        zonal_mean_diagnostics.data_vars[f"error-{example_output_var}"].attrs["units"]
        == "cm"
    )
    assert f"gen-{example_output_var}" in actual_var_names
    assert (
        zonal_mean_diagnostics.data_vars[f"gen-{example_output_var}"].attrs["units"]
        == ""
    )
    assert len(zonal_mean_diagnostics.coords) == 1
    assert "lat" in zonal_mean_diagnostics.coords

    for source in ["target", "prediction"]:
        histograms = xr.open_dataset(
            tmp_path / f"histograms_{source}.nc", decode_timedelta=False
        )
        actual_var_names = sorted([str(k) for k in histograms.keys()])
        # NOTE: target histograms include forcing variables
        n_vars = (
            len(all_out_names)
            if source == "prediction"
            else len(all_out_names) + len(forcing_names)
        )
        assert len(actual_var_names) == 2 * n_vars
        assert example_output_var in actual_var_names
        assert histograms.data_vars[example_output_var].attrs["units"] == "count"
        assert f"{example_output_var}_bin_edges" in actual_var_names
        assert (
            histograms.data_vars[f"{example_output_var}_bin_edges"].attrs["units"]
            == "cm"
        )
        var_counts_per_timestep = histograms[example_output_var].sum(dim=["bin"])
        same_count_each_timestep = np.all(
            var_counts_per_timestep.values == var_counts_per_timestep.values[0]
        )
        assert same_count_each_timestep
    if monthly_reference_filename is not None:
        assert f"inference/annual/{example_output_var}" in wandb_logs[-1]
        assert f"inference/annual/r2_gen_{example_output_var}" in wandb_logs[-1]
        assert f"inference/annual/r2_target_{example_output_var}" in wandb_logs[-1]
    assert "inference/total_steps_per_second" in wandb_logs[-1]


@pytest.mark.parametrize(
    "n_forward_steps,forward_steps_in_memory", [(10, 2), (10, 10), (15, 4)]
)
def test_inference_writer_boundaries(
    tmp_path: pathlib.Path, n_forward_steps: int, forward_steps_in_memory: int
):
    """Test that data at initial condition boundaires"""
    in_names = ["var"]
    out_names = ["var"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("lat", 4), DimSize("lon", 8)]

    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(log_to_screen=True, log_to_file=False, log_to_wandb=True),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            time_coarsen=None,
        ),
        forward_steps_in_memory=forward_steps_in_memory,
        allow_incompatible_dataset=True,  # stepper checkpoint has arbitrary info
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)
    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(
            yaml_config=str(config_filename),
        )
        inference_logs = wandb.get_logs()
    n_ic_timesteps = 1
    summary_log_step = 1
    assert (
        len(inference_logs)
        == n_ic_timesteps + config.n_forward_steps + summary_log_step
    )

    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    target_ds = xr.open_dataset(
        tmp_path / "autoregressive_target.nc", decode_timedelta=False
    )
    # data writers do not include initial condition
    assert len(prediction_ds["time"]) == n_forward_steps
    assert not np.any(np.isnan(prediction_ds["var"].values))

    gen = prediction_ds["var"]
    tar = target_ds["var"]
    gen_time_mean = torch.from_numpy(gen.mean(dim="time").values)
    tar_time_mean = torch.from_numpy(tar.mean(dim="time").values)
    area_weights = metrics.spherical_area_weights(
        tar["lat"].values, num_lon=len(tar["lon"])
    )
    # check time mean metrics
    tol = 1e-4  # relative tolerance
    assert metrics.root_mean_squared_error(
        tar_time_mean, gen_time_mean, area_weights
    ).item() == pytest.approx(
        inference_logs[-1]["inference/time_mean/rmse/var"], rel=tol
    )
    assert metrics.weighted_mean_bias(
        tar_time_mean, gen_time_mean, area_weights
    ).item() == pytest.approx(
        inference_logs[-1]["inference/time_mean/bias/var"], rel=tol
    )

    prediction_ds = prediction_ds.isel(sample=0)
    target_ds = target_ds.isel(sample=0)
    ds = xr.open_dataset(data.data_filename, decode_timedelta=False)

    for i in range(0, n_forward_steps):
        # metrics logs includes IC while saved data does not
        log = inference_logs[i + n_ic_timesteps]
        # metric steps should match lead times
        assert log["inference/mean/forecast_step"] == i + n_ic_timesteps
        gen_i = torch.from_numpy(gen.isel(time=i).values)
        tar_i = torch.from_numpy(tar.isel(time=i).values)
        # check that manually computed metrics match logged metrics
        assert metrics.root_mean_squared_error(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_rmse/var"], rel=tol)
        assert metrics.weighted_mean_bias(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_bias/var"], rel=tol)
        assert metrics.gradient_magnitude_percent_diff(
            tar_i, gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(
            log["inference/mean/weighted_grad_mag_percent_diff/var"], rel=tol
        )
        assert metrics.weighted_mean(
            gen_i, area_weights, dim=(-2, -1)
        ).item() == pytest.approx(log["inference/mean/weighted_mean_gen/var"], rel=tol)

        # the target obs should be the same as the validation data obs
        # ds is original data which includes IC, target_ds does not
        np.testing.assert_allclose(
            target_ds["var"].isel(time=i).values,
            ds["var"].isel(time=i + 1).values,
        )
        if i > 0:
            lead_da = prediction_ds["var"].isel(time=i)
            # predictions should be previous condition + 1
            np.testing.assert_allclose(
                lead_da.values,
                prediction_ds["var"].isel(time=i - 1).values + 1,
            )
            # prediction and target should not have entirely the same values at
            # any lead > 0
            assert not np.allclose(
                lead_da.values,
                target_ds["var"].isel(time=i).values,
            )


def test_inference_data_time_coarsening(tmp_path: pathlib.Path):
    forward_steps_in_memory = 4
    coarsen_factor = 2
    in_names = ["var"]
    out_names = ["var"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("lat", 16), DimSize("lon", 32)]

    dim_sizes = DimSizes(
        n_time=9,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=8,
        forward_steps_in_memory=forward_steps_in_memory,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=False,
        ),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(
            save_prediction_files=True,
            log_extended_video_netcdfs=True,
            time_coarsen=TimeCoarsenConfig(coarsen_factor=coarsen_factor),
            save_histogram_files=True,
        ),
        allow_incompatible_dataset=True,  # stepper checkpoint has arbitrary info
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)
    main(yaml_config=str(config_filename))
    # check that the outputs all have the intended time dimension size
    prediction_ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    n_coarsened_timesteps = config.n_forward_steps // coarsen_factor
    assert (
        len(prediction_ds["time"]) == n_coarsened_timesteps
    ), "raw predictions time dimension size"
    metric_ds = xr.open_dataset(
        tmp_path / "reduced_autoregressive_predictions.nc", decode_timedelta=False
    )
    assert (
        metric_ds.sizes["timestep"] == n_coarsened_timesteps
    ), "reduced predictions time dimension size"
    for source in ["target", "prediction"]:
        histograms = xr.open_dataset(
            tmp_path / f"histograms_{source}.nc", decode_timedelta=False
        )
        assert (
            histograms.sizes["time"] == n_coarsened_timesteps
        ), "histograms time dimension size"


@pytest.mark.parametrize("has_required_fields", [True, False])
def test_compute_derived_quantities(has_required_fields):
    """Checks that tensors are added to the data dictionary appropriately."""
    batch_size, n_ensemble, n_time, nx, ny, nz = 2, 3, 4, 5, 6, 7

    def _make_data():
        vars = ["a"]
        if has_required_fields:
            additional_fields = [f"specific_total_water_{i}" for i in range(nz)] + [
                "PRESsfc"
            ]
            vars += additional_fields
        return {
            var: torch.randn(
                batch_size, n_ensemble, n_time, nx, ny, device=get_device()
            )
            for var in vars
        }

    vertical_coordinate = HybridSigmaPressureCoordinate(
        ak=torch.linspace(0, 1, nz + 1, device=get_device()),
        bk=torch.linspace(0, 1, nz + 1, device=get_device()),
    )

    def derive_func(data: TensorMapping, forcing_data: TensorMapping) -> TensorDict:
        updated = compute_derived_quantities(
            dict(data),
            vertical_coordinate=vertical_coordinate,
            timestep=TIMESTEP,
            forcing_data=dict(forcing_data),
        )
        return updated

    metrics = {"loss": 42.0}
    gen_data = _make_data()
    target_data = {k: v[:, :1] for k, v in _make_data().items()}
    stepped = TrainOutput(
        metrics,
        EnsembleTensorDict(gen_data),
        EnsembleTensorDict(target_data),
        time=xr.DataArray(np.zeros((batch_size, n_time)), dims=["sample", "time"]),
        normalize=lambda x: x,
        derive_func=derive_func,
    )

    derived_stepped = stepped.compute_derived_variables()

    dry_air_name = "surface_pressure_due_to_dry_air"
    water_path_name = "total_water_path"
    existence_check = (
        dry_air_name in derived_stepped.gen_data
        and dry_air_name in derived_stepped.target_data
        and water_path_name in derived_stepped.gen_data
        and water_path_name in derived_stepped.target_data
    )

    if has_required_fields:
        assert existence_check
        for f in (
            derived_stepped.gen_data[dry_air_name],
            derived_stepped.gen_data["a"],
        ):
            assert f.shape == (batch_size, n_ensemble, n_time, nx, ny)
        for f in (
            derived_stepped.target_data[dry_air_name],
            derived_stepped.target_data["a"],
        ):
            assert f.shape == (batch_size, 1, n_time, nx, ny)
    else:
        assert not existence_check


def test_derived_metrics_run_without_errors(
    tmp_path: pathlib.Path, very_fast_only: bool
):
    """Checks that derived metrics are computed during inferece without errors."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")

    n_forward_steps = 2

    in_names = ["var", "PRESsfc", "specific_total_water_0", "specific_total_water_1"]
    out_names = ["var", "PRESsfc", "specific_total_water_0", "specific_total_water_1"]
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"

    horizontal = [DimSize("lat", 16), DimSize("lon", 32)]

    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=3,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
        nz_interface=dim_sizes.nz_interface,
    )
    time_varying_values = [float(i) for i in range(dim_sizes.n_time)]
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        time_varying_values=time_varying_values,
        timestep_days=TIMESTEP.total_seconds() / 86400,
        num_data_workers=2,
    )
    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(
            log_to_screen=True,
            log_to_file=False,
            log_to_wandb=True,
        ),
        loader=data.inference_data_loader_config,
        prediction_loader=None,
        data_writer=DataWriterConfig(save_prediction_files=True),
        forward_steps_in_memory=1,
        allow_incompatible_dataset=True,  # stepper checkpoint has arbitrary info
    )
    config_filename = tmp_path / "config.yaml"
    with open(config_filename, "w") as f:
        yaml.dump(dataclasses.asdict(config), f)

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        main(yaml_config=str(config_filename))
        inference_logs = wandb.get_logs()

    # derived variables should not have normalized metrics reported
    assert "inference/mean_norm/weighted_rmse/total_water_path" not in inference_logs[0]
    assert "inference/time_mean_norm/rmse/total_water_path" not in inference_logs[-1]

    ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc", decode_timedelta=False
    )
    assert "units" in ds["total_water_path"].attrs
    assert "long_name" in ds["total_water_path"].attrs


@pytest.mark.parametrize(
    "time_coarsen,n_forward_steps,forward_steps_in_memory",
    [
        pytest.param(3, 12, 4, id="not_multiple_of_forward_steps_in_memory"),
        pytest.param(-1, 12, 4, id="invalid_time_coarsen"),
        pytest.param(2, 5, 4, id="not_multiple_of_n_forward_steps"),
    ],
)
def test_inference_config_raises_incompatible_timesteps(
    time_coarsen, n_forward_steps, forward_steps_in_memory
):
    base_config_dict = dict(
        experiment_dir="./some_dir",
        n_forward_steps=n_forward_steps,
        checkpoint_path="./some_dir",
        logging=LoggingConfig(),
        loader=InferenceDataLoaderConfig(
            dataset=XarrayDataConfig(
                data_path="./some_data",
            ),
            start_indices=InferenceInitialConditionIndices(
                first=0, n_initial_conditions=1, interval=1
            ),
        ),
    )
    base_config_dict["forward_steps_in_memory"] = forward_steps_in_memory
    base_config_dict["data_writer"] = {"time_coarsen": {"coarsen_factor": time_coarsen}}
    with pytest.raises(ValueError):
        dacite.from_dict(
            data_class=InferenceEvaluatorConfig,
            data=base_config_dict,
            config=dacite.Config(strict=True),
        )


def test_inference_override(tmp_path: pathlib.Path):
    """Test that data at initial condition boundaires"""
    in_names = ["co2", "surface_temperature", "ocean_fraction"]
    fluxes = ["ULWRFtoa"]
    out_names = ["surface_temperature"] + fluxes
    all_names = list(set(in_names).union(out_names))
    stepper_path = tmp_path / "stepper"
    n_forward_steps = 8

    horizontal = [DimSize("lat", 4), DimSize("lon", 8)]
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
    )
    data = FV3GFSData(
        path=tmp_path,
        names=all_names,
        dim_sizes=dim_sizes,
        timestep_days=TIMESTEP.total_seconds() / 86400,
    )
    ocean_override = OceanConfig(
        surface_temperature_name="surface_temperature",
        ocean_fraction_name="ocean_fraction",
    )
    multi_call_override = MultiCallConfig(
        forcing_name="co2",
        forcing_multipliers={"_quartered_co2": 0.25},
        output_names=fluxes,
    )
    stepper_override = StepperOverrideConfig(
        ocean=ocean_override, multi_call=multi_call_override
    )

    config = InferenceEvaluatorConfig(
        experiment_dir=str(tmp_path),
        n_forward_steps=n_forward_steps,
        checkpoint_path=str(stepper_path),
        logging=LoggingConfig(log_to_screen=True, log_to_file=False, log_to_wandb=True),
        loader=data.inference_data_loader_config,
        data_writer=DataWriterConfig(save_prediction_files=True, time_coarsen=None),
        forward_steps_in_memory=4,
        stepper_override=stepper_override,
        allow_incompatible_dataset=True,  # stepper checkpoint has arbitrary info
    )
    stepper = config.load_stepper()
    validate_stepper_ocean(stepper, ocean_override)
    validate_stepper_multi_call(stepper, multi_call_override)

    stepper_config = config.load_stepper_config()
    validate_stepper_config(
        stepper_config,
        expected_ocean_config=ocean_override,
        expected_multi_call_config=multi_call_override,
    )


def validate_stepper_config(
    stepper_config: StepperConfig,
    expected_ocean_config: OceanConfig | None,
    expected_multi_call_config: MultiCallConfig | None,
):
    assert isinstance(stepper_config.step._step_config_instance, MultiCallStepConfig)
    assert isinstance(
        stepper_config.step._step_config_instance.wrapped_step._step_config_instance,
        SingleModuleStepConfig,
    )
    assert (
        stepper_config.step._step_config_instance.wrapped_step._step_config_instance.ocean
        == expected_ocean_config
    )
    assert (
        stepper_config.step._step_config_instance.config == expected_multi_call_config
    )


@pytest.mark.parametrize(
    "use_correct_timestep",
    [True, False],
)
def test_inference_timestep_mismatch_error(
    tmp_path: pathlib.Path, use_correct_timestep: bool
):
    """Test that inference with a model trained with a different timestep than
    the forcing data raises an error.
    """
    in_names = ["var"]
    out_names = ["var"]
    stepper_path = tmp_path / "stepper_test_data"

    horizontal = [DimSize("lat", 4), DimSize("lon", 8)]
    dim_sizes = DimSizes(n_time=8, horizontal=horizontal, nz_interface=2)
    std = 1.0
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=std,
        data_shape=dim_sizes.shape_nd,
        timestep=TIMESTEP,
    )
    use_prediction_data = False
    n_forward_steps = 2
    if use_correct_timestep:
        with pytest.raises(ValueError) as err:
            inference_helper(
                tmp_path,
                in_names,
                out_names,
                use_prediction_data,
                dim_sizes,
                n_forward_steps,
                stepper_path,
                timestep=TIMESTEP,
                allow_incompatible_dataset_info=False,
            )
        assert "timestep is not compatible" not in str(err.value)
    else:
        with pytest.raises(ValueError, match="timestep is not compatible"):
            inference_helper(
                tmp_path,
                in_names,
                out_names,
                use_prediction_data,
                dim_sizes,
                n_forward_steps,
                stepper_path,
                timestep=datetime.timedelta(days=20),
                allow_incompatible_dataset_info=False,
            )


def test_inference_includes_diagnostics(tmp_path: pathlib.Path):
    """Test that diagnostics are included in evaluator metrics and outputs."""
    # NOTE: size of in_names and out_names has to be the same here or the
    # PlusOne outputs won't have the right shape
    in_names = ["prog", "forcing_var", "DSWRFtoa"]
    out_names = ["prog", "ULWRFtoa", "USWRFtoa"]
    stepper_path = tmp_path / "stepper"
    horizontal = [DimSize("lat", 16), DimSize("lon", 32)]
    use_prediction_data = False
    n_forward_steps = 2
    dim_sizes = DimSizes(
        n_time=n_forward_steps + 1,
        horizontal=horizontal,
        nz_interface=4,
    )
    save_plus_one_stepper(
        stepper_path,
        in_names,
        out_names,
        mean=0.0,
        std=1.0,
        data_shape=dim_sizes.shape_nd,
        timestep=datetime.timedelta(days=20),
    )
    inference_helper(
        tmp_path,
        in_names,
        out_names,
        use_prediction_data,
        dim_sizes,
        n_forward_steps,
        stepper_path,
        save_monthly_files=False,  # requires timestep == 6h
        timestep=datetime.timedelta(days=20),
        derived_names=["net_energy_flux_toa_into_atmosphere"],
    )
    ds = xr.open_dataset(
        tmp_path / "autoregressive_predictions.nc",
        decode_timedelta=False,
        decode_times=False,
    )
    # prognostic in
    assert "prog" in ds
    # diags in
    assert "ULWRFtoa" in ds
    assert "USWRFtoa" in ds
    # derived in
    assert "net_energy_flux_toa_into_atmosphere" in ds
    # forcings not in
    assert "DSWRFtoa" not in ds
    assert "forcing_var" not in ds
    # assert only prognostic variables are in initial condition and restart files
    for filename in ["initial_condition.nc", "restart.nc"]:
        ds = xr.open_dataset(tmp_path / filename, decode_timedelta=False)
        assert "USWRFtoa" not in ds
        assert "forcing_var" not in ds
        assert "prog" in ds


@pytest.mark.parametrize(
    ["stepper_metadata", "dataset_metadata"],
    [
        pytest.param(
            {"foo": VariableMetadata("m", "first definition of foo")},
            {"foo": VariableMetadata("mm", "second definition of foo")},
            id="different_metadata",
        ),
        pytest.param(
            {},
            {"foo": VariableMetadata("mm", "second definition of foo")},
            id="datset_has_metadata",
        ),
        pytest.param({}, {}, id="neither_have_metadata"),
    ],
)
@patch(
    "fme.ace.inference.evaluator.get_default_variable_metadata",
    return_value={"foo": VariableMetadata("cm", "third definition of foo")},
)
@patch("fme.ace.inference.evaluator.get_derived_variable_metadata", return_value={})
def test_resolve_variable_metadata(
    mock_get_derived_variable_metadata,
    mock_get_default_variable_metadata,
    stepper_metadata,
    dataset_metadata,
    caplog,
):
    with caplog.at_level(logging.WARNING):
        variable_metadata = resolve_variable_metadata(
            dataset_metadata=dataset_metadata,
            stepper_metadata=stepper_metadata,
            stepper_all_names=["foo"],
        )
    if "foo" in stepper_metadata:
        assert "foo" not in caplog.text
        assert variable_metadata == {
            "foo": VariableMetadata("m", "first definition of foo")
        }
    elif "foo" in dataset_metadata:
        assert "foo" not in caplog.text
        assert variable_metadata == {
            "foo": VariableMetadata("mm", "second definition of foo")
        }
    else:
        assert "foo" in caplog.text
        assert variable_metadata == {
            "foo": VariableMetadata("cm", "third definition of foo")
        }



import abc
import logging
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from typing import Any, ClassVar

import cftime
import numpy as np
import torch
import xarray as xr
from matplotlib import pyplot as plt

from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.typing_ import TensorDict, TensorMapping

from ...plotting import plot_mean_and_samples

SAMPLE_DIM, TIME_DIM, LAT_DIM, LON_DIM = 0, 1, -2, -1

SEA_SURFACE_TEMPERATURE_NAMES = ["sst", "surface_temperature", "TS"]


class Region(abc.ABC):
    @property
    @abc.abstractmethod
    def regional_weights(self) -> torch.Tensor: ...


@dataclass
class LatLonRegion(Region):
    lat: torch.Tensor
    lon: torch.Tensor
    lat_bounds: tuple[float, float]
    lon_bounds: tuple[float, float]
    horizontal_dims: ClassVar[tuple[int, int]] = (LAT_DIM, LON_DIM)

    def __post_init__(self):
        lat_mask = (
            (self.lat >= self.lat_bounds[0]) & (self.lat <= self.lat_bounds[1])
        ).unsqueeze(self.horizontal_dims[1])
        lon_mask = (
            (self.lon >= self.lon_bounds[0]) & (self.lon <= self.lon_bounds[1])
        ).unsqueeze(self.horizontal_dims[0])
        self._regional_weights = torch.where(
            torch.logical_and(lat_mask, lon_mask), 1.0, 0.0
        )

    @property
    def regional_weights(self) -> torch.Tensor:
        return self._regional_weights


def compute_nino34_index_power_spectrum(
    data: xr.DataArray,
    fs=1,
    time_dim: int = TIME_DIM,
    sample_dim: int = SAMPLE_DIM,
):
    """
    Compute the power spectrum of the input data.

    Args:
        data: A tensor of size n_time_steps containing the data.
        fs: The sampling frequency, defaults to 1 sample per month.
        time_dim: Time dimension index
        sample_dim: Sample dimension index
    """
    if len(data.shape) == 2:
        uhat = np.fft.rfft(data, axis=time_dim)
        power = np.abs(uhat) ** 2
        power = power.mean(axis=sample_dim)
    else:
        raise ValueError(
            f"Expected indicies to be of shape (sample, time) when calculating \
                nino 3.4 index power spectrum, got {data.shape}"
        )
    n_samples = data.shape[time_dim]
    freqs = np.fft.rfftfreq(n_samples, d=1 / fs)
    freqs_per_year = freqs / fs * 12.0
    return freqs_per_year, power


class RegionalIndexAggregator:
    """Aggregator for computing a regional index, in this case a monthly- and area-
    weighted average of a variable over a region.

    Args:
        regional_weights: A tensor of weights for the region.
        regional_mean: A function that computes the regional mean of a variable,
            given weights.
        running_mean_n_months: The number of months to use for the running mean.
    """

    _sample_dim = SAMPLE_DIM
    _time_dim = TIME_DIM
    sea_surface_temperature_names = SEA_SURFACE_TEMPERATURE_NAMES

    def __init__(
        self,
        regional_weights: torch.Tensor,
        regional_mean: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        running_mean_n_months: int = 5,
    ):
        self._regional_weights = regional_weights
        self._regional_mean = regional_mean
        self._running_mean_n_months = running_mean_n_months
        self._raw_indices: TensorDict = {}
        self._raw_index_times: xr.DataArray | None = None
        self._calendar: str | None = None
        self._already_logged: list[str] = []

    def record_batch(self, time: xr.DataArray, data: TensorMapping) -> None:
        for sst_name in self.sea_surface_temperature_names:
            if sst_name not in data:
                if sst_name not in self._already_logged:
                    logging.info(
                        f"Variable {sst_name} not found in data. "
                        "Skipping Nino3.4 index computation for this variable."
                    )
                    self._already_logged.append(sst_name)
                continue
            regional_average = self._regional_mean(
                data[sst_name], self._regional_weights
            )
            if sst_name not in self._raw_indices:
                self._raw_indices[sst_name] = regional_average
            else:
                self._raw_indices[sst_name] = torch.cat(
                    [self._raw_indices[sst_name], regional_average],
                    dim=self._time_dim,
                )
        if self._raw_index_times is None:
            self._raw_index_times = time
        else:
            self._raw_index_times = xr.concat([self._raw_index_times, time], dim="time")
        if self._calendar is None and time.dt.calendar is not None:
            self._calendar = time.dt.calendar

    def get_indices(self) -> xr.Dataset:
        indices = {}
        for sst_name in self.sea_surface_temperature_names:
            if sst_name not in self._raw_indices:
                continue
            anomaly_index_values = anomalies_from_monthly_climo(
                self._raw_indices[sst_name], self._raw_index_times
            )
            time_averaged_index, unique_months = running_monthly_mean(
                anomaly_index_values,
                self._raw_index_times,
                n_months=self._running_mean_n_months,
            )
            gathered_index = self._get_gathered_index(
                time_averaged_index, unique_months.years, unique_months.months
            )
            if gathered_index is None:
                continue
            indices[sst_name] = gathered_index
        return xr.Dataset(indices)

    def _get_gathered_index(
        self, index: torch.Tensor, years: torch.Tensor, months: torch.Tensor
    ) -> xr.DataArray | None:
        dist = Distributed.get_instance()
        if dist.world_size > 1:
            gathered_index, gathered_years, gathered_months = (
                dist.gather_irregular(index),
                dist.gather_irregular(years),
                dist.gather_irregular(months),
            )
            if (
                gathered_index is None
                or gathered_years is None
                or gathered_months is None
            ):
                return None
        else:
            gathered_index, gathered_years, gathered_months = [index], [years], [months]
        return self._to_index_data_array(
            gathered_index, gathered_years, gathered_months
        )

    def _to_index_data_array(
        self,
        indices: list[torch.Tensor],
        years: list[torch.Tensor],
        months: list[torch.Tensor],
    ) -> xr.DataArray:
        index_data_arrays = []
        for index, year, month in zip(indices, years, months):
            time_coord = [  # approximate the middle of the month
                cftime.datetime(
                    single_year.item(),
                    single_month.item(),
                    15,
                    calendar=(
                        self._calendar if self._calendar is not None else "standard"
                    ),
                )
                for single_year, single_month in zip(year.cpu(), month.cpu())
            ]
            index_data_arrays.append(
                xr.DataArray(
                    data=index.cpu().numpy(),
                    dims=["sample", "time"],
                    coords={"time": time_coord},
                )
            )
        return xr.concat(index_data_arrays, dim="sample")

    def get_logs(self, label: str) -> dict[str, Any]:
        indices = self.get_indices()
        plots = {}
        for sst_name in self.sea_surface_temperature_names:
            if sst_name in indices and indices[sst_name].sizes["time"] > 1:
                fig, ax = plt.subplots(1, 1)
                index_plottable = indices[sst_name].assign_coords(
                    {"time": convert_cftime_to_datetime_coord(indices[sst_name].time)}
                )
                plot_mean_and_samples(
                    ax, index_plottable, "ensemble_mean", time_series_dim="time"
                )
                ax.set_title("Nino3.4 Index")
                ax.set_ylabel("K")
                ax.legend()
                fig.tight_layout()
                plots[f"{sst_name}_nino34_index"] = fig
        for sst_name in self.sea_surface_temperature_names:
            if (
                sst_name in indices
                and indices[sst_name].dropna("time").sizes["time"] > 1
            ):
                freq, power_spectrum = compute_nino34_index_power_spectrum(
                    indices.dropna("time")[sst_name]
                )
                fig, ax = plt.subplots(1, 1)
                ax.plot(freq, power_spectrum, label="predicted ensemble mean")
                ax.set_title("Power Spectrum of Nino3.4 Index")
                ax.set_xlabel("Frequency [cycles/year]")
                ax.set_ylabel("Power [K**2]")
                ax.set(yscale="log")
                ax.legend()
                fig.tight_layout()
                plots[f"{sst_name}_nino34_index_power_spectrum"] = fig

        if len(label) > 0:
            label = label + "/"
        return {f"{label}{k}": v for k, v in plots.items()}

    def get_dataset(self) -> xr.Dataset:
        indices = self.get_indices()
        if indices is None:
            return xr.Dataset()
        else:
            return indices


class PairedRegionalIndexAggregator:
    """A paired (target, prediction) RegionalIndexAggregator."""

    def __init__(
        self,
        target_aggregator: RegionalIndexAggregator,
        prediction_aggregator: RegionalIndexAggregator,
    ):
        self._prediction_aggregator = prediction_aggregator
        self._target_aggregator = target_aggregator

    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ) -> None:
        self._target_aggregator.record_batch(time=time, data=target_data)
        self._prediction_aggregator.record_batch(time=time, data=gen_data)

    def get_logs(self, label: str) -> dict[str, Any]:
        target_indices = self._target_aggregator.get_indices()
        prediction_indices = self._prediction_aggregator.get_indices()
        plots = {}
        for sst_name in self._prediction_aggregator.sea_surface_temperature_names:
            if (
                sst_name in prediction_indices
                and prediction_indices[sst_name].sizes["time"] > 1
            ):
                target_indices_plottable = target_indices.assign_coords(
                    {"time": convert_cftime_to_datetime_coord(target_indices.time)}
                )
                prediction_indices_plottable = prediction_indices.assign_coords(
                    {"time": convert_cftime_to_datetime_coord(prediction_indices.time)}
                )
                fig, ax = plt.subplots(1, 1)
                plot_mean_and_samples(
                    ax,
                    prediction_indices_plottable[sst_name],
                    "predicted ensemble mean",
                    time_series_dim="time",
                )
                plot_mean_and_samples(
                    ax,
                    target_indices_plottable[sst_name],
                    "target",
                    time_series_dim="time",
                    color="orange",
                    plot_samples=False,
                )
                ax.set_title("Nino3.4 Index")
                ax.set_ylabel("K")
                ax.legend()
                fig.tight_layout()
                plots[f"{sst_name}_nino34_index"] = fig
        for sst_name in self._prediction_aggregator.sea_surface_temperature_names:
            if (
                sst_name in prediction_indices
                and prediction_indices[sst_name].dropna("time").sizes["time"] > 1
            ):
                freq, prediction_power_spectrum = compute_nino34_index_power_spectrum(
                    prediction_indices.dropna("time")[sst_name]
                )
                freq, target_power_spectrum = compute_nino34_index_power_spectrum(
                    target_indices.dropna("time")[sst_name]
                )
                fig, ax = plt.subplots(1, 1)
                ax.plot(
                    freq, prediction_power_spectrum, label="predicted ensemble mean"
                )
                ax.plot(freq, target_power_spectrum, label="target", color="orange")
                ax.set_title("Power Spectrum of Nino3.4 Index")
                ax.set_xlabel("Frequency [cycles/year]")
                ax.set(yscale="log")
                ax.legend()
                fig.tight_layout()
                plots[f"{sst_name}_nino34_index_power_spectrum"] = fig
        if len(label) > 0:
            label = label + "/"

        return {f"{label}{k}": v for k, v in plots.items()}

    def get_dataset(self) -> xr.Dataset:
        prediction = self._prediction_aggregator.get_dataset()
        target = self._target_aggregator.get_dataset()
        if len(prediction) == 0 or len(target) == 0:
            return xr.Dataset()
        else:
            return xr.concat(
                [
                    target.expand_dims({"source": ["target"]}),
                    prediction.expand_dims({"source": ["prediction"]}),
                ],
                dim="source",
            )


def anomalies_from_monthly_climo(
    data: torch.Tensor,
    time: xr.DataArray,
    time_dim: int = TIME_DIM,
) -> torch.Tensor:
    anomalies = torch.empty_like(data, device=data.device)
    for month in range(1, 13):
        month_mask = torch.tensor(
            time.dt.month.values == month, dtype=torch.bool, device=data.device
        )
        masked_data = data.where(
            month_mask, torch.tensor(float("nan"), device=data.device)
        )
        monthly_climo = (
            masked_data.nansum(dim=time_dim) / month_mask.sum(dim=time_dim)
        ).unsqueeze(dim=time_dim)
        anomalies = torch.where(month_mask, data - monthly_climo, anomalies)
    return anomalies


@dataclass
class UniqueMonths:
    years: torch.Tensor
    months: torch.Tensor

    @classmethod
    def from_arrays(cls, years: np.ndarray, months: np.ndarray) -> "UniqueMonths":
        if years.shape != months.shape:
            raise ValueError("Years and months arrays must have the same shape.")
        if years.ndim != 1:
            raise ValueError("Years and months arrays must be 1D.")
        unique_months_arr = np.unique(
            np.array(
                [(year, month) for year, month in zip(years, months)],
                dtype=("i,i"),
            ).astype("object")
        )
        return cls(
            years=torch.tensor(
                [unique_month[0] for unique_month in unique_months_arr],
                dtype=torch.int64,
                device=get_device(),
            ),
            months=torch.tensor(
                [unique_month[1] for unique_month in unique_months_arr],
                dtype=torch.int64,
                device=get_device(),
            ),
        )

    def __iter__(self):
        return zip(self.years, self.months)

    def __len__(self):
        return len(self.years)


def running_monthly_mean(
    data: torch.Tensor,
    time: xr.DataArray,
    n_months: int,
    time_dim: int = TIME_DIM,
    sample_dim: int = SAMPLE_DIM,
) -> tuple[torch.Tensor, UniqueMonths]:
    """Compute an n-month running mean of the input data.

    First compute the monthly mean of the data, then compute the running mean.
    """
    unique_months = UniqueMonths.from_arrays(
        time.dt.year.values.flatten(),
        time.dt.month.values.flatten(),
    )
    monthly_data = torch.full(
        (data.shape[sample_dim], len(unique_months)),
        fill_value=float("nan"),
        device=data.device,
    )
    running_monthly_data = torch.full_like(monthly_data, fill_value=float("nan"))
    for i_month, (year, month) in enumerate(unique_months):
        month_mask = torch.tensor(
            (time.dt.year.values == year.item())
            & (time.dt.month.values == month.item()),
            dtype=torch.bool,
            device=data.device,
        )
        masked_data = data.where(
            month_mask, torch.tensor(float("nan"), device=data.device)
        )
        monthly_data[:, i_month] = masked_data.nanmean(dim=TIME_DIM)
        if i_month >= n_months - 1:
            running_monthly_data[:, i_month] = monthly_data[
                :, i_month - n_months + 1 : i_month + 1
            ].nanmean(dim=time_dim)
    return running_monthly_data, unique_months


def convert_cftime_to_datetime_coord(time_coord: xr.DataArray) -> xr.DataArray:
    """Convert cftime array to datetime array that is approximately the same.
    We do this because it's necessary to plot data via plotly with datetime coordinates
    in order to get ticklabels that are intelligible, but we want to maintain a cftime
    coordinate for accuracy and saving raw data.
    """
    return xr.DataArray(
        data=[
            datetime(time.item().year, time.item().month, time.item().day)
            for time in time_coord
        ],
        dims=time_coord.dims,
    )



"""This module provides an entrypoint for benchmarking data loading performance.
It uses a time.sleep() to mock ML training but allows loading real datasets.
"""

import argparse
import dataclasses
import logging
import os
import shutil
import time
import uuid

import dacite
import yaml

from fme.core import logging_utils
from fme.core.dicts import to_flat_dict
from fme.core.timing import GlobalTimer
from fme.core.wandb import WandB

from ..requirements import DataRequirements
from .config import DataLoaderConfig
from .getters import get_gridded_data

TMPDIR = f"/tmp/ace_benchmark_{uuid.uuid4()}"


@dataclasses.dataclass
class BenchmarkConfig:
    """Configuration for benchmarking data loading."""

    loader: DataLoaderConfig
    logging: logging_utils.LoggingConfig
    names: list[str]
    n_timesteps: int
    train: bool = True
    sleep: float = 0.1

    def build(self):
        return get_gridded_data(
            self.loader,
            train=self.train,
            requirements=DataRequirements(self.names, self.n_timesteps),
        )

    def configure_wandb(self, env_vars: dict | None = None, **kwargs):
        config = to_flat_dict(dataclasses.asdict(self))
        # our wandb class requires "experiment_dir" to be in config
        config["experiment_dir"] = TMPDIR
        os.makedirs(TMPDIR)
        self.logging.configure_wandb(config=config, env_vars=env_vars, **kwargs)

    def configure_logging(self):
        self.logging.configure_logging("/tmp", "log.txt")


def benchmark(config: BenchmarkConfig):
    config.configure_logging()
    env_vars = logging_utils.retrieve_env_vars()
    beaker_url = logging_utils.log_beaker_url()
    config.configure_wandb(env_vars=env_vars, notes=beaker_url)
    wandb = WandB.get_instance()

    with GlobalTimer():
        logging.info("Initializing data loader.")
        timer = GlobalTimer.get_instance()
        with timer.context("initialization"):
            data = config.build()
            loader = data.loader

        logging.info("Getting data size in bytes from first batch.")
        example_batch = next(iter(loader))
        bytes_per_batch = sum(
            [x.numel() * x.element_size() for x in example_batch.data.values()]
        )
        logging.info(f"Each batch will be {bytes_per_batch / 1e6:.3f} MB.")
        if config.sleep > 0:
            # assuming no time wasted data loading, estimate max throughput
            max_throughput = bytes_per_batch / config.sleep
            logging.info(
                "Max throughput possible given batch size and configured sleep: "
                f"{max_throughput / 1e6:.2f} MB/s"
            )
        logging.info(f"Starting loop to load {len(loader)} batches.")
        timer.start("data_loading")
        seconds_per_batch = time.time()
        for i, batch in enumerate(loader):
            timer.stop()
            if i % 10 == 0:
                logging.info(f"Loaded batch {i}")
            with timer.context("sleeping"):
                time.sleep(config.sleep)
            seconds_per_batch = time.time() - seconds_per_batch
            wandb.log({"seconds_per_batch": seconds_per_batch}, step=i)
            seconds_per_batch = time.time()
            timer.start("data_loading")
        timer.stop()
        logging.info(f"Finished loading {len(loader)} batches.")
        total_time = timer.get_duration("data_loading") + timer.get_duration("sleeping")
        actual_throughput = (bytes_per_batch * len(loader)) / total_time
        logging.info(f"Actual throughput achieved: {actual_throughput / 1e6:.2f} MB/s")
        logging.info("Timer results:")
        logging.info(timer.get_durations())
        wandb_logs = timer.get_durations() | {
            "total_time": total_time,
            "throughput_mb_per_s": actual_throughput / 1e6,
            "mb_per_batch": bytes_per_batch / 1e6,
        }
        if config.sleep > 0:
            ratio = actual_throughput / max_throughput
            wandb_logs["ratio_of_actual_to_possible_throughput"] = ratio
            wandb_logs["max_possible_throughput_mb_per_s"] = max_throughput / 1e6
        wandb.log(wandb_logs, step=i)
        shutil.rmtree(TMPDIR)


def main():
    parser = argparse.ArgumentParser(description="Benchmark ACE data loading.")
    parser.add_argument("config", help="Path to the configuration file.")
    args = parser.parse_args()

    with open(args.config) as f:
        config = dacite.from_dict(
            BenchmarkConfig, yaml.safe_load(f), config=dacite.Config(strict=True)
        )

    benchmark(config)


if __name__ == "__main__":
    main()



import dataclasses
import re
import tempfile

import numpy as np
import pytest
import torch
import xarray as xr

from fme.ace.inference.evaluator import main as inference_evaluator_main
from fme.ace.testing import DimSizes, MonthlyReferenceData, save_scalar_netcdf
from fme.ace.testing.fv3gfs_data import get_nd_dataset
from fme.ace.train.train import main as train_main
from fme.core.coordinates import LatLonCoordinates
from fme.core.testing.wandb import mock_wandb


def save_ocean_nd_netcdf(
    filename: str,
    dim_sizes: DimSizes,
    variable_names: list[str],
    timestep_days: float = 1.0,
    nz_levels: int = 2,  # Number of ocean levels for thetao_0, mask_0 etc.
):
    """
    Saves a netCDF file with synthetic ocean data, including masks and NaNs.
    Inspired by fme.ace.testing.fv3gfs_data.save_nd_netcdf.
    """
    base_variable_names = [
        name
        for name in variable_names
        if not name.startswith("mask_") and not name.startswith("idepth_")
    ]

    # Use a modified DimSizes for base dataset to avoid issues with nz_interface
    # if get_nd_dataset interprets it in a way not suitable for ocean vars
    base_dim_sizes = dataclasses.replace(dim_sizes, nz_interface=1)

    ds = get_nd_dataset(
        dim_sizes=base_dim_sizes,
        variable_names=base_variable_names,
        timestep_days=timestep_days,
        include_vertical_coordinate=False,  # No ak/bk for ocean
    )

    rng = np.random.default_rng(0)  # For reproducibility

    horizontal_dims = list(map(lambda x: x.name, dim_sizes.horizontal))
    horizontal_shape = list(map(lambda x: x.size, dim_sizes.horizontal))

    masks_data = {}
    for i in range(nz_levels):
        mask_name = f"mask_{i}"
        if mask_name in variable_names:
            mask_data_values = rng.integers(low=0, high=2, size=horizontal_shape)
            coords_for_mask = {
                dim: ds.coords[dim] for dim in horizontal_dims if dim in ds.coords
            }
            masks_data[i] = xr.DataArray(
                mask_data_values,
                dims=horizontal_dims,
                coords=coords_for_mask,
                name=mask_name,
            )
            ds[mask_name] = masks_data[i]

    if "mask_0" in ds:
        ds["mask_2d"] = ds["mask_0"].copy()

    for var_name in base_variable_names:
        if var_name not in ds:  # Skip if not generated by get_nd_dataset
            continue
        match = re.fullmatch(r"(.+)_(\d+)", var_name)
        if match:
            level = int(match.group(2))
            if level < nz_levels and level in masks_data:
                ds[var_name] = ds[var_name].where(masks_data[level] == 1, float("nan"))
        else:  # Assume 2D variable, apply mask_0 if it exists (e.g., for sst)
            if 0 in masks_data and "mask_0" in variable_names:
                ds[var_name] = ds[var_name].where(masks_data[0] == 1, float("nan"))

    # Add idepth variables (interface depths)
    # nz_interface in DimSizes should be nz_levels + 1
    if dim_sizes.nz_interface != nz_levels + 1:
        raise ValueError(
            f"dim_sizes.nz_interface ({dim_sizes.nz_interface}) "
            f"must be nz_levels ({nz_levels}) + 1."
        )
    for i in range(dim_sizes.nz_interface):
        idepth_name = f"idepth_{i}"
        if idepth_name in variable_names:
            ds[idepth_name] = float(i)  # Saved as 0-D variable

    ds.to_netcdf(filename, unlimited_dims=["time"])
    return ds


_TRAIN_CONFIG_TEMPLATE = """
experiment_dir: {results_dir}
save_checkpoint: true
save_per_epoch_diagnostics: {save_per_epoch_diagnostics}
max_epochs: {max_epochs}
n_forward_steps: 2
logging:
  log_to_screen: true
  log_to_wandb: {log_to_wandb}
  log_to_file: false
  project: fme
  entity: ai2cm
train_loader:
  dataset:
    - data_path: '{train_data_path}'
      spatial_dimensions: latlon
  batch_size: 2
  num_data_workers: 0
validation_loader:
  dataset:
    - data_path: '{valid_data_path}'
      spatial_dimensions: latlon
  batch_size: 2
  num_data_workers: 0
optimization:
  use_gradient_accumulation: true
  optimizer_type: "Adam"
  lr: 0.001
  kwargs:
    weight_decay: 0.01
  scheduler:
      type: CosineAnnealingLR
      kwargs:
        T_max: 1
stepper:
  loss:
    type: "MSE"
  input_masking:
    mask_value: 0
    fill_value: 0.0
  step:
    type: single_module
    config:
      in_names: {in_variable_names}
      out_names: {out_variable_names}
      normalization:
        network:
          global_means_path: '{global_means_path}'
          global_stds_path: '{global_stds_path}'
      builder:
        type: Samudra
        config:
          ch_width: [8, 16]
          dilation: [2, 4]
          n_layers: [1, 1]
          norm: batch
          norm_kwargs:
            track_running_stats: false
      corrector:
        type: "ocean_corrector"
        config:
          sea_ice_fraction_correction:
            sea_ice_fraction_name: sea_ice_fraction
            land_fraction_name: land_fraction
          ocean_heat_content_correction: true
inference:
  aggregator:
    monthly_reference_data: {monthly_data_filename}
  loader:
    dataset:
      data_path: '{valid_data_path}'
      spatial_dimensions: latlon
    start_indices:
      first: 0
      n_initial_conditions: 2
      interval: 1
  n_forward_steps: {inference_forward_steps}
  forward_steps_in_memory: 2
"""

_INFERENCE_CONFIG_TEMPLATE = """
experiment_dir: {results_dir}
n_forward_steps: 6
forward_steps_in_memory: 2
checkpoint_path: {results_dir}/training_checkpoints/best_ckpt.tar
data_writer:
  save_prediction_files: true
aggregator:
  log_video: true
logging:
  log_to_screen: true
  log_to_wandb: {log_to_wandb}
  log_to_file: false
  project: fme
  entity: ai2cm
loader:
  dataset:
    data_path: '{valid_data_path}'
    spatial_dimensions: latlon
    fill_nans:
      method: constant
      value: 0.0
  start_indices:
    first: 0
    n_initial_conditions: 2
    interval: 1
"""


def _get_test_yaml_files(
    train_data_path,
    valid_data_path,
    monthly_data_filename,
    results_dir,
    global_means_path,
    global_stds_path,
    in_variable_names,
    out_variable_names,
    log_to_wandb=False,
    max_epochs=1,
    inference_forward_steps=2,
    save_per_epoch_diagnostics=False,
):
    train_config_content = _TRAIN_CONFIG_TEMPLATE

    train_string = train_config_content.format(
        train_data_path=train_data_path,
        valid_data_path=valid_data_path,
        monthly_data_filename=monthly_data_filename,
        results_dir=results_dir,
        global_means_path=global_means_path,
        global_stds_path=global_stds_path,
        in_variable_names=in_variable_names,
        out_variable_names=out_variable_names,
        log_to_wandb=str(log_to_wandb).lower(),
        max_epochs=max_epochs,
        inference_forward_steps=inference_forward_steps,
        save_per_epoch_diagnostics=str(save_per_epoch_diagnostics).lower(),
    )

    inference_string = _INFERENCE_CONFIG_TEMPLATE.format(
        results_dir=results_dir,
        log_to_wandb=str(log_to_wandb).lower(),
        valid_data_path=valid_data_path,  # Inference uses validation data path
    )

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".yaml") as f_train:
        f_train.write(train_string)

    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, suffix=".yaml"
    ) as f_inference:
        f_inference.write(inference_string)

    return f_train.name, f_inference.name


def get_ocean_dim_sizes(
    n_lat=16,
    n_lon=32,
    n_time=3,
    nz_levels=2,  # For thetao_0, thetao_1, mask_0, mask_1
) -> DimSizes:
    return DimSizes(
        n_time=n_time,
        horizontal=LatLonCoordinates(
            lon=torch.arange(float(n_lon)),
            lat=torch.arange(float(n_lat)),
        ).loaded_sizes,
        nz_interface=nz_levels + 1,  # For idepth_0, idepth_1, idepth_2
    )


def _setup(
    path,
    log_to_wandb=True,
    max_epochs=1,
    n_time=60,
    timestep_days=2,
    inference_forward_steps=10,
    save_per_epoch_diagnostics=True,
    nz_levels=2,
):
    if not path.exists():
        path.mkdir(parents=True, exist_ok=True)
    seed = 0
    np.random.seed(seed)
    torch.manual_seed(seed)

    # Ocean-specific variable names
    in_variable_names = [
        "thetao_0",
        "thetao_1",
        "sst",
        "hfds",
        "hfgeou",
        "sea_surface_fraction",
        "sea_ice_fraction",
        "land_fraction",
    ]
    out_variable_names = [
        "thetao_0",
        "thetao_1",
        "sst",
        "sea_ice_fraction",
    ]

    # Add masks and idepths for data generation
    mask_names = [f"mask_{i}" for i in range(nz_levels)]
    idepth_names = [f"idepth_{i}" for i in range(nz_levels + 1)]

    in_variable_names_for_data_gen = list(set(in_variable_names + mask_names))

    all_variable_names_for_data_gen = list(
        set(in_variable_names_for_data_gen + out_variable_names + idepth_names)
    )
    stats_variable_names = list(set(in_variable_names + out_variable_names))

    dim_sizes = get_ocean_dim_sizes(n_time=n_time, nz_levels=nz_levels)

    data_dir = path / "data"
    stats_dir = path / "stats"
    results_dir = path / "results"
    data_dir.mkdir(exist_ok=True)
    stats_dir.mkdir(exist_ok=True)
    results_dir.mkdir(exist_ok=True)

    save_ocean_nd_netcdf(
        filename=data_dir / "data.nc",
        dim_sizes=dim_sizes,
        variable_names=all_variable_names_for_data_gen,
        timestep_days=timestep_days,
        nz_levels=nz_levels,
    )
    save_scalar_netcdf(
        stats_dir / "stats-mean.nc",
        variable_names=stats_variable_names,
    )
    save_scalar_netcdf(
        stats_dir / "stats-stddev.nc",
        variable_names=stats_variable_names,
    )

    monthly_dim_sizes = get_ocean_dim_sizes(n_time=10 * 12, nz_levels=1)
    monthly_out_vars = [
        v for v in out_variable_names if not re.match(r".+_\d+", v) or v.endswith("_0")
    ]
    if not monthly_out_vars and out_variable_names:
        monthly_out_vars = [out_variable_names[0]]

    if monthly_out_vars:
        monthly_reference_data = MonthlyReferenceData(
            path=data_dir,
            names=monthly_out_vars,
            dim_sizes=monthly_dim_sizes,
            n_ensemble=3,
        )
        monthly_data_filename = monthly_reference_data.data_filename
    else:
        monthly_data_filename = "null"

    train_config_filename, inference_config_filename = _get_test_yaml_files(
        train_data_path=data_dir,
        valid_data_path=data_dir,
        monthly_data_filename=monthly_data_filename,
        results_dir=results_dir,
        global_means_path=stats_dir / "stats-mean.nc",
        global_stds_path=stats_dir / "stats-stddev.nc",
        in_variable_names=in_variable_names,
        out_variable_names=out_variable_names,
        log_to_wandb=log_to_wandb,
        max_epochs=max_epochs,
        inference_forward_steps=inference_forward_steps,
        save_per_epoch_diagnostics=save_per_epoch_diagnostics,
    )
    return train_config_filename, inference_config_filename


def test_train_and_inference(tmp_path, very_fast_only: bool):
    """Ensure that ACE Ocean training and subsequent standalone inference run."""
    if very_fast_only:
        pytest.skip("Skipping non-fast tests")

    train_config, inference_config = _setup(tmp_path)

    with mock_wandb() as wandb:
        train_main(yaml_config=train_config)
        wandb_logs = wandb.get_logs()
        for log in wandb_logs:
            assert "inference/mean/forecast_step" not in log

    results_dir = tmp_path / "results"

    validation_output_dir = results_dir / "output" / "val" / "epoch_0001"
    assert validation_output_dir.exists()
    validation_diags = ["mean"]
    validation_map_diags = ["snapshot", "mean_map"]
    for diagnostic in validation_diags + validation_map_diags:
        diagnostic_output = validation_output_dir / f"{diagnostic}_diagnostics.nc"
        assert diagnostic_output.exists()
        ds = xr.open_dataset(diagnostic_output, decode_timedelta=False)
        assert len(ds) > 0
        for var in ds.data_vars:
            if diagnostic == "mean":
                assert not np.isnan(ds[var].values).any()
            elif "ocean_heat_content" in var:
                assert not np.isnan(ds[var].values).all()
                assert np.isnan(ds[var].values).any()
            else:
                # other
                assert not np.isnan(ds[var].values).all()
                assert np.isnan(ds[var].values).any()

    inline_inference_output_dir = results_dir / "output" / "inference" / "epoch_0001"
    assert inline_inference_output_dir.exists()
    for diagnostic in ("time_mean", "time_mean_norm"):
        diagnostic_output = inline_inference_output_dir / f"{diagnostic}_diagnostics.nc"
        assert diagnostic_output.exists()
        ds = xr.open_dataset(diagnostic_output, decode_timedelta=False)
        assert len(ds) > 0
        for var in ds.data_vars:
            assert not np.isnan(ds[var].values).all()
            assert np.isnan(ds[var].values).any()

    with mock_wandb() as wandb:
        wandb.configure(log_to_wandb=True)
        inference_evaluator_main(yaml_config=inference_config)
        wandb_logs = wandb.get_logs()

    tm_logs = wandb_logs[-1]
    for name, metric in tm_logs.items():
        if "power_spectrum" in name:
            # power spectra not well defined for masked data
            continue
        if isinstance(metric, float):
            assert not np.isnan(metric), f"{name} should not be NaN"

    prediction_output_path = results_dir / "autoregressive_predictions.nc"
    best_checkpoint_path = results_dir / "training_checkpoints" / "best_ckpt.tar"
    assert prediction_output_path.exists()
    assert best_checkpoint_path.exists()

    ds_prediction = xr.open_dataset(prediction_output_path, decode_timedelta=False)
    for name in ["sst", "thetao_0", "thetao_1", "ocean_heat_content"]:
        assert name in ds_prediction
        # outputs should have some non-null values
        assert not np.isnan(ds_prediction[name].values).all()
        # outputs should have some masked regions
        assert np.isnan(ds_prediction[name].values).any()

    ds_target = xr.open_dataset(
        results_dir / "autoregressive_target.nc", decode_timedelta=False
    )
    assert "sst" in ds_target

    for diagnostic in ("mean", "mean_norm"):
        diagnostic_output = results_dir / f"{diagnostic}_diagnostics.nc"
        assert diagnostic_output.exists()
        ds = xr.open_dataset(diagnostic_output, decode_timedelta=False)
        assert len(ds) > 0
        for var in ds.data_vars:
            assert not np.isnan(ds[var].values).any()



import os
import tempfile

import numpy as np
import torch
import xarray as xr

from fme.ace.data_loading.batch_data import BatchData
from fme.ace.inference.data_writer.dataset_metadata import DatasetMetadata
from fme.ace.inference.data_writer.main import _write
from fme.core.dataset.data_typing import VariableMetadata


def test_write_single_timestep():
    n_samples = 2
    n_lat = 4
    n_lon = 5
    n_time = 1
    batch = BatchData.new_on_cpu(
        data={"air_temperature": torch.rand((n_samples, n_time, n_lat, n_lon))},
        time=xr.DataArray(np.random.rand(n_samples, n_time), dims=["sample", "time"]),
        horizontal_dims=["lat", "lon"],
    )
    with tempfile.TemporaryDirectory() as tmpdir:
        _write(
            data=batch,
            path=tmpdir,
            filename="initial_condition.nc",
            variable_metadata={
                "air_temperature": VariableMetadata(
                    long_name="Air Temperature", units="K"
                )
            },
            coords={"lat": np.arange(n_lat), "lon": np.arange(n_lon)},
            dataset_metadata=DatasetMetadata(
                history={"created": "2023-10-01T00:00:00"}
            ),
        )
        filename = os.path.join(tmpdir, "initial_condition.nc")
        assert os.path.exists(filename)
        with xr.open_dataset(filename, decode_timedelta=False) as ds:
            assert "air_temperature" in ds
            assert ds.air_temperature.shape == (n_samples, n_lat, n_lon)
            assert ds.time.shape == (n_samples,)
            assert ds.air_temperature.dims == ("sample", "lat", "lon")
            xr.testing.assert_allclose(ds.time, batch.time.isel(time=0))
            np.testing.assert_allclose(
                ds.air_temperature.values,
                batch.data["air_temperature"].squeeze(dim=1).cpu().numpy(),
            )
            np.testing.assert_allclose(ds.coords["lat"].values, np.arange(n_lat))
            np.testing.assert_allclose(ds.coords["lon"].values, np.arange(n_lon))
            assert ds.air_temperature.attrs["long_name"] == "Air Temperature"
            assert ds.air_temperature.attrs["units"] == "K"
            assert ds.attrs["history.created"] == "2023-10-01T00:00:00"


def test_write_multiple_timesteps():
    n_samples = 2
    n_lat = 4
    n_lon = 5
    n_time = 2
    batch = BatchData.new_on_cpu(
        data={"air_temperature": torch.rand((n_samples, n_time, n_lat, n_lon))},
        time=xr.DataArray(np.random.rand(n_samples, n_time), dims=["sample", "time"]),
        horizontal_dims=["lat", "lon"],
    )
    with tempfile.TemporaryDirectory() as tmpdir:
        _write(
            data=batch,
            path=tmpdir,
            filename="initial_condition.nc",
            variable_metadata={
                "air_temperature": VariableMetadata(
                    long_name="Air Temperature", units="K"
                )
            },
            coords={"lat": np.arange(n_lat), "lon": np.arange(n_lon)},
            dataset_metadata=DatasetMetadata(),
        )
        filename = os.path.join(tmpdir, "initial_condition.nc")
        assert os.path.exists(filename)
        with xr.open_dataset(filename, decode_timedelta=False) as ds:
            assert "air_temperature" in ds
            assert ds.air_temperature.shape == (n_samples, n_time, n_lat, n_lon)
            assert ds.time.shape == (n_samples, n_time)
            assert ds.air_temperature.dims == ("sample", "time", "lat", "lon")
            np.testing.assert_allclose(ds.time.values, batch.time.values)
            np.testing.assert_allclose(
                ds.air_temperature.values, batch.data["air_temperature"].cpu().numpy()
            )
            np.testing.assert_allclose(ds.coords["lat"].values, np.arange(n_lat))
            np.testing.assert_allclose(ds.coords["lon"].values, np.arange(n_lon))
            assert ds.air_temperature.attrs["long_name"] == "Air Temperature"
            assert ds.air_temperature.attrs["units"] == "K"



# This module is derived from the train.py module in the following repository:
# https://github.com/NVlabs/FourCastNet. The corresponding license is
# provided below.

# BSD 3-Clause License
#
# Copyright (c) 2022, FourCastNet authors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The code was authored by the following people:
#
# Jaideep Pathak - NVIDIA Corporation
# Shashank Subramanian - NERSC, Lawrence Berkeley National Laboratory
# Peter Harrington - NERSC, Lawrence Berkeley National Laboratory
# Sanjeev Raja - NERSC, Lawrence Berkeley National Laboratory
# Ashesh Chattopadhyay - Rice University
# Morteza Mardani - NVIDIA Corporation
# Thorsten Kurth - NVIDIA Corporation
# David Hall - NVIDIA Corporation
# Zongyi Li - California Institute of Technology, NVIDIA Corporation
# Kamyar Azizzadenesheli - Purdue University
# Pedram Hassanzadeh - Rice University
# Karthik Kashinath - NVIDIA Corporation
# Animashree Anandkumar - California Institute of Technology, NVIDIA Corporation

import abc
import contextlib
import gc
import logging
import os
import time
import uuid
from collections.abc import Callable, Mapping
from typing import Any, ClassVar, Generic, Protocol, TypeVar

import torch

import fme
from fme.core.distributed import Distributed
from fme.core.ema import EMATracker
from fme.core.generics.aggregator import AggregatorABC, InferenceAggregatorABC
from fme.core.generics.data import GriddedDataABC, InferenceDataABC
from fme.core.generics.inference import run_inference
from fme.core.generics.train_stepper import TrainOutputABC, TrainStepperABC
from fme.core.optimization import NullOptimization, Optimization
from fme.core.timing import GlobalTimer
from fme.core.training_history import TrainingJob
from fme.core.typing_ import Slice
from fme.core.wandb import WandB


class EndOfBatchCallback(Protocol):
    def __call__(self) -> None: ...


class EndOfEpochCallback(Protocol):
    def __call__(self, epoch: int) -> Mapping[str, Any]: ...


def null_end_of_epoch_callback(epoch: int) -> Mapping[str, Any]:
    return {}


class TrainConfigProtocol(Protocol):
    __dataclass_fields__: ClassVar[dict[str, Any]]

    @property
    def experiment_dir(self) -> str: ...

    @property
    def output_dir(self) -> str: ...

    @property
    def checkpoint_dir(self) -> str: ...

    @property
    def max_epochs(self) -> int: ...

    @property
    def save_checkpoint(self) -> bool: ...

    @property
    def validate_using_ema(self) -> bool: ...

    @property
    def log_train_every_n_batches(self) -> int: ...

    @property
    def segment_epochs(self) -> int | None: ...

    @property
    def checkpoint_save_epochs(self) -> Slice | None: ...

    @property
    def ema_checkpoint_save_epochs(self) -> Slice | None: ...

    @property
    def evaluate_before_training(self) -> bool: ...

    def get_inference_epochs(self) -> list[int]: ...


PS = TypeVar("PS", contravariant=True)  # prognostic state
TO = TypeVar("TO", bound="TrainOutputABC")  # train output
BD = TypeVar("BD")  # batch data for training
FD = TypeVar("FD")  # forcing data for inference
SD = TypeVar("SD")  # stepped data from inference


class AggregatorBuilderABC(abc.ABC, Generic[PS, TO, SD]):
    @abc.abstractmethod
    def get_train_aggregator(self) -> AggregatorABC[TO]:
        pass

    @abc.abstractmethod
    def get_validation_aggregator(self) -> AggregatorABC[TO]:
        pass

    @abc.abstractmethod
    def get_inference_aggregator(self) -> InferenceAggregatorABC[PS, SD]:
        pass


class CheckpointPaths:
    def __init__(self, checkpoint_dir: str):
        self.checkpoint_dir = checkpoint_dir

    @property
    def latest_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "ckpt.tar")

    @property
    def best_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "best_ckpt.tar")

    @property
    def best_inference_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "best_inference_ckpt.tar")

    @property
    def ema_checkpoint_path(self) -> str:
        return os.path.join(self.checkpoint_dir, "ema_ckpt.tar")

    def epoch_checkpoint_path(self, epoch: int) -> str:
        return os.path.join(self.checkpoint_dir, f"ckpt_{epoch:04d}.tar")

    def ema_epoch_checkpoint_path(self, epoch: int) -> str:
        return os.path.join(self.checkpoint_dir, f"ema_ckpt_{epoch:04d}.tar")


class Trainer:
    def __init__(
        self,
        train_data: GriddedDataABC[BD],
        validation_data: GriddedDataABC[BD],
        inference_data: InferenceDataABC[PS, FD],
        stepper: TrainStepperABC[PS, BD, FD, SD, TO],
        build_optimization: Callable[[torch.nn.ModuleList], Optimization],
        build_ema: Callable[[torch.nn.ModuleList], EMATracker],
        config: TrainConfigProtocol,
        aggregator_builder: AggregatorBuilderABC[PS, TO, SD],
        end_of_batch_callback: EndOfBatchCallback = lambda: None,
        end_of_epoch_callback: EndOfEpochCallback = null_end_of_epoch_callback,
        do_gc_collect: bool = True,
    ):
        logging.info(f"Current device is {fme.get_device()}")
        self.dist = Distributed.get_instance()
        if self.dist.is_root():
            if not os.path.isdir(config.experiment_dir):
                os.makedirs(config.experiment_dir)
            if not os.path.isdir(config.checkpoint_dir):
                os.makedirs(config.checkpoint_dir)
        self.config = config
        self.paths = CheckpointPaths(config.checkpoint_dir)

        self.train_data = train_data
        self.valid_data = validation_data
        for gridded_data, name in zip(
            (self.train_data, self.valid_data), ("train", "valid")
        ):
            gridded_data.log_info(name)

        self.num_batches_seen = 0
        self._start_epoch = 0
        self._epoch = self._start_epoch
        self.num_batches_seen = 0
        self._best_validation_loss = torch.inf
        self._best_inference_error = torch.inf

        self.stepper = stepper
        self.stepper.update_training_history(TrainingJob.from_env())

        self.optimization = build_optimization(stepper.modules)
        self._end_of_batch_callback = end_of_batch_callback
        self._end_of_epoch_callback = end_of_epoch_callback
        self._no_optimization = NullOptimization()
        self._aggregator_builder = aggregator_builder

        resuming = os.path.isfile(self.paths.latest_checkpoint_path)
        if resuming:
            logging.info(f"Resuming training from {self.paths.latest_checkpoint_path}")
            self.restore_checkpoint(
                self.paths.latest_checkpoint_path, self.paths.ema_checkpoint_path
            )

        wandb = WandB.get_instance()
        wandb.watch(self.stepper.modules)

        n_params = count_parameters(self.stepper.modules)
        logging.info(f"Number of trainable model parameters: {n_params}")

        self._inference_data = inference_data
        self._ema = build_ema(stepper.modules)
        self._do_gc_collect = do_gc_collect

    def set_end_of_epoch_callback(self, end_of_epoch_callback: EndOfEpochCallback):
        self._end_of_epoch_callback = end_of_epoch_callback

    def switch_off_grad(self, model: torch.nn.Module):
        for param in model.parameters():
            param.requires_grad = False

    def train(self):
        logging.info("Starting Training Loop...")

        self._epoch = self._start_epoch
        inference_epochs = self.config.get_inference_epochs()
        if self.config.segment_epochs is None:
            segment_max_epochs = self.config.max_epochs
        else:
            segment_max_epochs = min(
                self._start_epoch + self.config.segment_epochs, self.config.max_epochs
            )

        if self.config.evaluate_before_training and self._epoch == 0:
            logging.info("Starting validation before training")
            valid_logs = self.validate_one_epoch()
            if self._epoch in inference_epochs:
                logging.info("Starting inline inference before training")
                inference_logs = self.inference_one_epoch()
            else:
                inference_logs = {}
            valid_loss = valid_logs["val/mean/loss"]
            logging.info(f"Validation loss before training: {valid_loss}")
            logging.info("Logging to wandb")
            all_logs = valid_logs | inference_logs | {"epoch": self._epoch}
            wandb = WandB.get_instance()
            wandb.log(all_logs, step=self.num_batches_seen)

        while self._epoch < segment_max_epochs:
            self._epoch += 1
            if self._do_gc_collect:
                # garbage collect to avoid CUDA error in some contexts
                # https://github.com/pytorch/pytorch/issues/67978#issuecomment-1661986812  # noqa: E501
                gc.collect()
            logging.info(f"Epoch: {self._epoch}")
            self.train_data.set_epoch(self._epoch)

            start_time = time.time()
            logging.info(f"Starting training step on epoch {self._epoch}")
            train_logs = self.train_one_epoch()
            train_end = time.time()
            logging.info(f"Starting validation step on epoch {self._epoch}")
            valid_logs = self.validate_one_epoch()
            valid_end = time.time()
            if self._epoch in inference_epochs:
                logging.info(f"Starting inference step on epoch {self._epoch}")
                inference_logs = self.inference_one_epoch()
                inference_end: float | None = time.time()
            else:
                inference_logs = {}
                inference_end = None

            train_loss = train_logs["train/mean/loss"]
            valid_loss = valid_logs["val/mean/loss"]
            inference_error = inference_logs.get(
                "inference/time_mean_norm/rmse/channel_mean", None
            )
            # need to get the learning rate before stepping the scheduler
            lr = self.optimization.learning_rate
            self.optimization.step_scheduler(valid_loss)

            if self.dist.is_root():
                if self.config.save_checkpoint:
                    logging.info(f"Saving checkpoints for epoch {self._epoch}")
                    self.save_all_checkpoints(valid_loss, inference_error)

            time_elapsed = time.time() - start_time
            logging.info(f"Time taken for epoch {self._epoch} is {time_elapsed} sec")
            logging.info(f"Train loss: {train_loss}. Valid loss: {valid_loss}")
            if inference_error is not None:
                logging.info(f"Inference error: {inference_error}")

            with self.validation_context():
                additional_logs = self._end_of_epoch_callback(self._epoch)

            logging.info("Logging to wandb")
            all_logs = {
                **train_logs,
                **valid_logs,
                **inference_logs,
                **additional_logs,
                **{
                    "lr": lr,
                    "epoch": self._epoch,
                    "epoch_train_seconds": train_end - start_time,
                    "epoch_validation_seconds": valid_end - train_end,
                    "epoch_total_seconds": time_elapsed,
                    "best_val_loss": self._best_validation_loss,
                    "best_inference_error": self._best_inference_error,
                },
            }
            if inference_end is not None:
                all_logs["epoch_inference_seconds"] = inference_end - valid_end
            wandb = WandB.get_instance()
            wandb.log(all_logs, step=self.num_batches_seen)

    def train_one_epoch(self):
        """Train for one epoch and return logs from TrainAggregator."""
        wandb = WandB.get_instance()
        names_to_log = ("batch_loss", "training_samples_per_second_on_rank_0")
        aggregator = self._aggregator_builder.get_train_aggregator()
        n_samples_seen_since_logging = 0
        self.stepper.set_train()
        if self.num_batches_seen == 0:
            # Before training, log the loss on the first batch.
            with torch.no_grad(), GlobalTimer():
                batch = next(iter(self.train_data.loader))
                stepped = self.stepper.train_on_batch(
                    batch,
                    optimization=self._no_optimization,
                )

                if self.config.log_train_every_n_batches > 0:
                    with torch.no_grad():
                        metrics = {
                            f"batch_{name}": self.dist.reduce_mean(metric)
                            for name, metric in sorted(stepped.get_metrics().items())
                        }
                    wandb.log(metrics, step=self.num_batches_seen)
        current_time = time.time()
        for batch in self.train_data.loader:
            with GlobalTimer():
                stepped = self.stepper.train_on_batch(batch, self.optimization)
            aggregator.record_batch(stepped)
            self._end_of_batch_callback()
            self._ema(model=self.stepper.modules)
            self.num_batches_seen += 1
            n_samples_seen_since_logging += self.train_data.batch_size
            if (
                self.config.log_train_every_n_batches > 0
                and self.num_batches_seen % self.config.log_train_every_n_batches == 0
            ):
                with torch.no_grad():
                    metrics = {
                        f"batch_{name}": self.dist.reduce_mean(metric)
                        for name, metric in sorted(stepped.get_metrics().items())
                    }
                duration = time.time() - current_time
                current_time = time.time()
                samples_per_second = n_samples_seen_since_logging / duration
                metrics["training_samples_per_second_on_rank_0"] = samples_per_second
                wandb.log(metrics, step=self.num_batches_seen)
                metrics_to_log = {k: metrics[k] for k in names_to_log if k in metrics}
                logging.info(f"Step {self.num_batches_seen}: {metrics_to_log}")
                n_samples_seen_since_logging = 0
        aggregator.flush_diagnostics(subdir=f"epoch_{self._epoch:04d}")
        return aggregator.get_logs(label="train")

    @contextlib.contextmanager
    def validation_context(self):
        """
        The context for running validation.

        In this context, the stepper uses the EMA model if
        `self.config.validate_using_ema` is True.
        """
        if self.config.validate_using_ema:
            with self._ema_context():
                yield
        else:
            yield

    @contextlib.contextmanager
    def _ema_context(self):
        """
        A context where the stepper uses the EMA model.
        """
        self._ema.store(parameters=self.stepper.modules.parameters())
        self._ema.copy_to(model=self.stepper.modules)
        try:
            yield
        finally:
            self._ema.restore(parameters=self.stepper.modules.parameters())

    def validate_one_epoch(self):
        self.stepper.set_eval()
        aggregator = self._aggregator_builder.get_validation_aggregator()
        logging.info("Starting loop over validation data")
        with torch.no_grad(), self.validation_context(), GlobalTimer():
            for batch in self.valid_data.loader:
                stepped = self.stepper.train_on_batch(
                    batch,
                    optimization=NullOptimization(),
                    compute_derived_variables=True,
                )
                aggregator.record_batch(
                    batch=stepped,
                )
        logging.info("Starting flush of reduced diagnostics to disk")
        aggregator.flush_diagnostics(subdir=f"epoch_{self._epoch:04d}")
        logging.info("Getting validation aggregator logs")
        return aggregator.get_logs(label="val")

    def inference_one_epoch(
        self,
    ):
        logging.info("Starting inline inference run")
        return inference_one_epoch(
            stepper=self.stepper,
            validation_context=self.validation_context,
            dataset=self._inference_data,
            aggregator=self._aggregator_builder.get_inference_aggregator(),
            label="inference",
            epoch=self._epoch,
        )

    def save_checkpoint(self, checkpoint_path, include_optimization=False):
        # save to a temporary file in case we get pre-empted during save
        temporary_location = os.path.join(
            os.path.dirname(checkpoint_path), f".{uuid.uuid4()}.tmp"
        )
        try:
            data = {
                "num_batches_seen": self.num_batches_seen,
                "epoch": self._epoch,
                "best_validation_loss": self._best_validation_loss,
                "best_inference_error": self._best_inference_error,
                "stepper": self.stepper.get_state(),
                "ema": self._ema.get_state(),
            }
            if include_optimization:
                data["optimization"] = self.optimization.get_state()
            torch.save(data, temporary_location)
            os.replace(temporary_location, checkpoint_path)
        finally:
            if os.path.exists(temporary_location):
                os.remove(temporary_location)

    def restore_checkpoint(self, checkpoint_path, ema_checkpoint_path):
        """
        Restore the checkpoint from the given path. This includes the existing state of
        the stepper, optimization, training epoch, and EMA. This is most suitable
        for resuming training from a checkpoint without changing the training schedule,
        i.e., to manage preemption.
        """
        _restore_checkpoint(self, checkpoint_path, ema_checkpoint_path)

    def _epoch_checkpoint_enabled(self, epoch: int) -> bool:
        return epoch_checkpoint_enabled(
            epoch, self.config.max_epochs, self.config.checkpoint_save_epochs
        )

    def _ema_epoch_checkpoint_enabled(self, epoch: int) -> bool:
        return epoch_checkpoint_enabled(
            epoch, self.config.max_epochs, self.config.ema_checkpoint_save_epochs
        )

    def save_all_checkpoints(self, valid_loss: float, inference_error: float | None):
        if self.config.validate_using_ema:
            best_checkpoint_context = self._ema_context
        else:
            best_checkpoint_context = contextlib.nullcontext  # type: ignore
        with best_checkpoint_context():
            save_best_checkpoint = False
            if valid_loss <= self._best_validation_loss:
                logging.info(
                    "Saving lowest validation loss checkpoint to "
                    f"{self.paths.best_checkpoint_path}"
                )
                self._best_validation_loss = valid_loss
                save_best_checkpoint = True  # wait until inference error is updated
            if inference_error is not None and (
                inference_error <= self._best_inference_error
            ):
                logging.info(
                    f"Epoch inference error ({inference_error}) is lower than "
                    f"previous best inference error ({self._best_inference_error})."
                )
                logging.info(
                    "Saving lowest inference error checkpoint to "
                    f"{self.paths.best_inference_checkpoint_path}"
                )
                self._best_inference_error = inference_error
                self.save_checkpoint(self.paths.best_inference_checkpoint_path)
            if save_best_checkpoint:
                self.save_checkpoint(self.paths.best_checkpoint_path)

        logging.info(f"Saving latest checkpoint to {self.paths.latest_checkpoint_path}")
        self.save_checkpoint(
            self.paths.latest_checkpoint_path, include_optimization=True
        )
        with self._ema_context():
            logging.info(
                f"Saving latest EMA checkpoint to {self.paths.ema_checkpoint_path}"
            )
            self.save_checkpoint(self.paths.ema_checkpoint_path)
        if self._epoch_checkpoint_enabled(self._epoch):
            epoch_checkpoint_path = self.paths.epoch_checkpoint_path(self._epoch)
            logging.info(f"Saving epoch checkpoint to {epoch_checkpoint_path}")
            self.save_checkpoint(epoch_checkpoint_path)
        if self._ema_epoch_checkpoint_enabled(self._epoch):
            ema_epoch_checkpoint_path = self.paths.ema_epoch_checkpoint_path(
                self._epoch
            )
            logging.info(f"Saving EMA epoch checkpoint to {ema_epoch_checkpoint_path}")
            with self._ema_context():
                self.save_checkpoint(ema_epoch_checkpoint_path)


def inference_one_epoch(
    stepper: TrainStepperABC[PS, BD, FD, SD, TO],
    validation_context: Callable[[], contextlib.AbstractContextManager],
    dataset: InferenceDataABC[PS, FD],
    aggregator: InferenceAggregatorABC[PS, SD],
    label: str,
    epoch: int,
):
    stepper.set_eval()
    with torch.no_grad(), validation_context(), GlobalTimer():
        run_inference(
            predict=stepper.predict_paired,
            data=dataset,
            aggregator=aggregator,
        )
    logging.info("Starting flush of reduced diagnostics to disk")
    aggregator.flush_diagnostics(subdir=f"epoch_{epoch:04d}")
    logging.info("Getting inline inference aggregator logs")
    logs = aggregator.get_summary_logs()
    return {f"{label}/{k}": v for k, v in logs.items()}


def _restore_checkpoint(trainer: Trainer, checkpoint_path, ema_checkpoint_path):
    checkpoint = torch.load(
        checkpoint_path, map_location=fme.get_device(), weights_only=False
    )
    trainer.stepper.load_state(checkpoint["stepper"])
    trainer.optimization.load_state(checkpoint["optimization"])
    trainer.num_batches_seen = checkpoint["num_batches_seen"]
    trainer._start_epoch = checkpoint["epoch"]
    trainer._best_validation_loss = checkpoint["best_validation_loss"]
    trainer._best_inference_error = checkpoint["best_inference_error"]
    ema_checkpoint = torch.load(
        ema_checkpoint_path, map_location=fme.get_device(), weights_only=False
    )
    ema_stepper: TrainStepperABC = type(trainer.stepper).from_state(
        ema_checkpoint["stepper"]
    )
    trainer._ema = EMATracker.from_state(checkpoint["ema"], ema_stepper.modules)


def count_parameters(modules: torch.nn.ModuleList) -> int:
    parameters = 0
    for module in modules:
        for parameter in module.parameters():
            if parameter.requires_grad:
                parameters += parameter.numel()
    return parameters


def epoch_checkpoint_enabled(
    epoch: int, max_epochs: int, save_epochs: Slice | None
) -> bool:
    if save_epochs is None:
        return False
    return epoch in range(max_epochs)[save_epochs.slice]



# flake8: noqa
# Copied from tag v0.1.0 https://github.com/NVIDIA/modulus-makani/tree/0218658e925a3c88c6513022a33ecc1647735c2e
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import tensorly as tl
import torch

tl.set_backend("pytorch")

from tltorch.factorized_tensors.core import FactorizedTensor

from .contractions import (
    _contract_dhconv,
    _contract_dhconv_real,
    _contract_diagonal,
    _contract_diagonal_real,
    _contract_sep_dhconv,
    _contract_sep_dhconv_real,
    _contract_sep_diagonal,
    _contract_sep_diagonal_real,
)

einsum_symbols = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


def _contract_dense(x, weight, separable=False, operator_type="diagonal"):
    order = tl.ndim(x)
    # batch-size, in_channels, x, y...
    x_syms = list(einsum_symbols[:order])

    # in_channels, out_channels, x, y...
    weight_syms = list(x_syms[1:])  # no batch-size

    # batch-size, out_channels, x, y...
    if separable:
        out_syms = [x_syms[0]] + list(weight_syms)
    else:
        weight_syms.insert(1, einsum_symbols[order])  # outputs
        out_syms = list(weight_syms)
        out_syms[0] = x_syms[0]

    if operator_type == "diagonal":
        pass
    elif operator_type == "block-diagonal":
        weight_syms.insert(-1, einsum_symbols[order + 1])
        out_syms[-1] = weight_syms[-2]
    elif operator_type == "dhconv":
        weight_syms.pop()
    else:
        raise ValueError(f"Unkonw operator type {operator_type}")

    eq = "".join(x_syms) + "," + "".join(weight_syms) + "->" + "".join(out_syms)

    if not torch.is_tensor(weight):
        weight = weight.to_tensor()

    res = tl.einsum(eq, x, weight).contiguous()

    return res


def _contract_cp(x, cp_weight, separable=False, operator_type="diagonal"):
    order = tl.ndim(x)

    x_syms = str(einsum_symbols[:order])
    rank_sym = einsum_symbols[order]
    out_sym = einsum_symbols[order + 1]
    out_syms = list(x_syms)

    if separable:
        factor_syms = [einsum_symbols[1] + rank_sym]  # in only
    else:
        out_syms[1] = out_sym
        factor_syms = [einsum_symbols[1] + rank_sym, out_sym + rank_sym]  # in, out

    factor_syms += [xs + rank_sym for xs in x_syms[2:]]  # x, y, ...

    if operator_type == "diagonal":
        pass
    elif operator_type == "block-diagonal":
        out_syms[-1] = einsum_symbols[order + 2]
        factor_syms += [out_syms[-1] + rank_sym]
    elif operator_type == "dhconv":
        factor_syms.pop()
    else:
        raise ValueError(f"Unkonw operator type {operator_type}")

    eq = (
        x_syms + "," + rank_sym + "," + ",".join(factor_syms) + "->" + "".join(out_syms)
    )

    res = tl.einsum(eq, x, cp_weight.weights, *cp_weight.factors).contiguous()

    return res


def _contract_tucker(x, tucker_weight, separable=False, operator_type="diagonal"):
    order = tl.ndim(x)

    x_syms = str(einsum_symbols[:order])
    out_sym = einsum_symbols[order]
    out_syms = list(x_syms)
    if separable:
        core_syms = einsum_symbols[order + 1 : 2 * order]
        factor_syms = [xs + rs for (xs, rs) in zip(x_syms[1:], core_syms)]  # x, y, ...

    else:
        core_syms = einsum_symbols[order + 1 : 2 * order + 1]
        out_syms[1] = out_sym
        factor_syms = [
            einsum_symbols[1] + core_syms[0],
            out_sym + core_syms[1],
        ]  # out, in
        factor_syms += [
            xs + rs for (xs, rs) in zip(x_syms[2:], core_syms[2:])
        ]  # x, y, ...

    if operator_type == "diagonal":
        pass
    elif operator_type == "block-diagonal":
        raise NotImplementedError(
            f"Operator type {operator_type} not implemented for Tucker"
        )
    else:
        raise ValueError(f"Unkonw operator type {operator_type}")

    eq = (
        x_syms
        + ","
        + core_syms
        + ","
        + ",".join(factor_syms)
        + "->"
        + "".join(out_syms)
    )

    res = tl.einsum(eq, x, tucker_weight.core, *tucker_weight.factors).contiguous()

    return res


def _contract_tt(x, tt_weight, separable=False, operator_type="diagonal"):
    order = tl.ndim(x)

    x_syms = list(einsum_symbols[:order])
    weight_syms = list(x_syms[1:])  # no batch-size

    if not separable:
        weight_syms.insert(1, einsum_symbols[order])  # outputs
        out_syms = list(weight_syms)
        out_syms[0] = x_syms[0]
    else:
        out_syms = list(x_syms)

    if operator_type == "diagonal":
        pass
    elif operator_type == "block-diagonal":
        weight_syms.insert(-1, einsum_symbols[order + 1])
        out_syms[-1] = weight_syms[-2]
    elif operator_type == "dhconv":
        weight_syms.pop()
    else:
        raise ValueError(f"Unkonw operator type {operator_type}")

    rank_syms = list(einsum_symbols[order + 2 :])
    tt_syms = []
    for i, s in enumerate(weight_syms):
        tt_syms.append([rank_syms[i], s, rank_syms[i + 1]])
    eq = (
        "".join(x_syms)
        + ","
        + ",".join("".join(f) for f in tt_syms)
        + "->"
        + "".join(out_syms)
    )

    res = tl.einsum(eq, x, *tt_weight.factors).contiguous()

    return res


# jitted PyTorch contractions:
def _contract_dense_pytorch(
    x, weight, separable=False, operator_type="diagonal", complex=True
):
    # make sure input is contig
    x = x.contiguous()

    if separable:
        if operator_type == "diagonal":
            if complex:
                x = _contract_sep_diagonal(x, weight)
            else:
                x = _contract_sep_diagonal_real(x, weight)
        elif operator_type == "dhconv":
            if complex:
                x = _contract_sep_dhconv(x, weight)
            else:
                x = _contract_sep_dhconv_real(x, weight)
        else:
            raise ValueError(f"Unkonw operator type {operator_type}")
    else:
        if operator_type == "diagonal":
            if complex:
                x = _contract_diagonal(x, weight)
            else:
                x = _contract_diagonal_real(x, weight)
        elif operator_type == "dhconv":
            if complex:
                x = _contract_dhconv(x, weight)
            else:
                x = _contract_dhconv_real(x, weight)
        else:
            raise ValueError(f"Unkonw operator type {operator_type}")

    # make contiguous
    x = x.contiguous()
    return x


def _contract_dense_reconstruct(
    x, weight, separable=False, operator_type="diagonal", complex=True
):
    """Contraction for dense tensors, factorized or not"""
    if not torch.is_tensor(weight):
        weight = weight.to_tensor()
        # weight = torch.view_as_real(weight)

    return _contract_dense_pytorch(
        x, weight, separable=separable, operator_type=operator_type, complex=complex
    )


def get_contract_fun(
    weight,
    implementation="reconstructed",
    separable=False,
    operator_type="diagonal",
    complex=True,
):
    """Generic ND implementation of Fourier Spectral Conv contraction

    Parameters
    ----------
    weight : tensorly-torch's FactorizedTensor
    implementation : {'reconstructed', 'factorized'}, default is 'reconstructed'
        whether to reconstruct the weight and do a forward pass (reconstructed)
        or contract directly the factors of the factorized weight with the input (factorized)

    Returns
    -------
    function : (x, weight) -> x * weight in Fourier space
    """
    if implementation == "reconstructed":
        handle = partial(
            _contract_dense_reconstruct,
            separable=separable,
            complex=complex,
            operator_type=operator_type,
        )
        return handle
    elif implementation == "factorized":
        if torch.is_tensor(weight):
            handle = partial(
                _contract_dense_pytorch,
                separable=separable,
                complex=complex,
                operator_type=operator_type,
            )
            return handle
        elif isinstance(weight, FactorizedTensor):
            if weight.name.lower() == "complexdense" or weight.name.lower() == "dense":
                return _contract_dense
            elif weight.name.lower() == "complextucker":
                return _contract_tucker
            elif weight.name.lower() == "complextt":
                return _contract_tt
            elif weight.name.lower() == "complexcp":
                return _contract_cp
            else:
                raise ValueError(f"Got unexpected factorized weight type {weight.name}")
        else:
            raise ValueError(
                f"Got unexpected weight type of class {weight.__class__.__name__}"
            )
    else:
        raise ValueError(
            f'Got {implementation=}, expected "reconstructed" or "factorized"'
        )



from .dataset_metadata import DatasetMetadata


def test_dataset_metadata_as_flat_str_dict():
    dataset_metadata = DatasetMetadata(
        source={
            "inference_git_sha": "abc123",
            "inference_fme_version": "1.0",
            "additional_info": {"some_other_key": "some_other_value"},
        },
        history={"created": "2023-10-01T00:00:00"},
        title="Test Dataset",
    )
    flat_dict = dataset_metadata.as_flat_str_dict()
    assert flat_dict == {
        "source.inference_git_sha": "abc123",
        "source.inference_fme_version": "1.0",
        "source.additional_info.some_other_key": "some_other_value",
        "history.created": "2023-10-01T00:00:00",
        "title": "Test Dataset",
    }



# flake8: noqa
# Copied from tag v0.1.0 https://github.com/NVIDIA/modulus-makani/tree/0218658e925a3c88c6513022a33ecc1647735c2e
# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import torch
import torch.fft
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint


@torch.jit.script
def drop_path(
    x: torch.Tensor, drop_prob: float = 0.0, training: bool = False
) -> torch.Tensor:
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1.0 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2d ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class PatchEmbed(nn.Module):
    def __init__(
        self, img_size=(224, 224), patch_size=(16, 16), in_chans=3, embed_dim=768
    ):
        super(PatchEmbed, self).__init__()
        self.red_img_size = (
            (img_size[0] // patch_size[0]),
            (img_size[1] // patch_size[1]),
        )
        num_patches = self.red_img_size[0] * self.red_img_size[1]
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches
        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size, bias=True
        )
        self.proj.weight.is_shared_mp = ["spatial"]
        self.proj.bias.is_shared_mp = ["spatial"]

    def forward(self, x):
        # gather input
        B, C, H, W = x.shape
        assert (
            H == self.img_size[0] and W == self.img_size[1]
        ), f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        # new: B, C, H*W
        x = self.proj(x).flatten(2)
        return x


class EncoderDecoder(nn.Module):
    def __init__(
        self,
        num_layers,
        input_dim,
        output_dim,
        hidden_dim,
        act_layer,
        gain=1.0,
        input_format="nchw",
    ):
        super(EncoderDecoder, self).__init__()

        encoder_modules = []
        current_dim = input_dim
        for i in range(num_layers):
            # fully connected layer
            if input_format == "nchw":
                encoder_modules.append(nn.Conv2d(current_dim, hidden_dim, 1, bias=True))
            elif input_format == "traditional":
                encoder_modules.append(nn.Linear(current_dim, hidden_dim, bias=True))
            else:
                raise NotImplementedError(
                    f"Error, input format {input_format} not supported."
                )

            # weight sharing
            encoder_modules[-1].weight.is_shared_mp = ["spatial"]

            # proper initializaiton
            scale = math.sqrt(2.0 / current_dim)
            nn.init.normal_(encoder_modules[-1].weight, mean=0.0, std=scale)
            if encoder_modules[-1].bias is not None:
                encoder_modules[-1].bias.is_shared_mp = ["spatial"]
                nn.init.constant_(encoder_modules[-1].bias, 0.0)

            encoder_modules.append(act_layer())
            current_dim = hidden_dim

        # final output layer
        if input_format == "nchw":
            encoder_modules.append(nn.Conv2d(current_dim, output_dim, 1, bias=False))
        elif input_format == "traditional":
            encoder_modules.append(nn.Linear(current_dim, output_dim, bias=False))

        # weight sharing
        encoder_modules[-1].weight.is_shared_mp = ["spatial"]

        # proper initializaiton
        scale = math.sqrt(gain / current_dim)
        nn.init.normal_(encoder_modules[-1].weight, mean=0.0, std=scale)
        if encoder_modules[-1].bias is not None:
            encoder_modules[-1].bias.is_shared_mp = ["spatial"]
            nn.init.constant_(encoder_modules[-1].bias, 0.0)

        self.fwd = nn.Sequential(*encoder_modules)

    def forward(self, x):
        return self.fwd(x)


class MLP(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        output_bias=True,
        input_format="nchw",
        drop_rate=0.0,
        drop_type="iid",
        checkpointing=0,
        gain=1.0,
        **kwargs,
    ):
        super(MLP, self).__init__()
        self.checkpointing = checkpointing
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features

        # First fully connected layer
        if input_format == "nchw":
            fc1 = nn.Conv2d(in_features, hidden_features, 1, bias=True)
            fc1.weight.is_shared_mp = ["spatial"]
            fc1.bias.is_shared_mp = ["spatial"]
        elif input_format == "traditional":
            fc1 = nn.Linear(in_features, hidden_features, bias=True)
        else:
            raise NotImplementedError(
                f"Error, input format {input_format} not supported."
            )

        # initialize the weights correctly
        scale = math.sqrt(2.0 / in_features)
        nn.init.normal_(fc1.weight, mean=0.0, std=scale)
        nn.init.constant_(fc1.bias, 0.0)

        # activation
        act = act_layer()

        # sanity checks
        if (input_format == "traditional") and (drop_type == "features"):
            raise NotImplementedError(
                f"Error, traditional input format and feature dropout cannot be selected simultaneously"
            )

        # output layer
        if input_format == "nchw":
            fc2 = nn.Conv2d(hidden_features, out_features, 1, bias=output_bias)
            fc2.weight.is_shared_mp = ["spatial"]
            if output_bias:
                fc2.bias.is_shared_mp = ["spatial"]
        elif input_format == "traditional":
            fc2 = nn.Linear(hidden_features, out_features, bias=output_bias)
        else:
            raise NotImplementedError(
                f"Error, input format {input_format} not supported."
            )

        # gain factor for the output determines the scaling of the output init
        scale = math.sqrt(gain / hidden_features)
        nn.init.normal_(fc2.weight, mean=0.0, std=scale)
        if fc2.bias is not None:
            nn.init.constant_(fc2.bias, 0.0)

        if drop_rate > 0.0:
            if drop_type == "iid":
                drop = nn.Dropout(drop_rate)
            elif drop_type == "features":
                drop = nn.Dropout2d(drop_rate)
            else:
                raise NotImplementedError(f"Error, drop_type {drop_type} not supported")
        else:
            drop = nn.Identity()

        # create forward pass
        self.fwd = nn.Sequential(fc1, act, drop, fc2, drop)

    @torch.jit.ignore
    def checkpoint_forward(self, x):
        return checkpoint(self.fwd, x, use_reentrant=False)

    def forward(self, x):
        if self.checkpointing >= 2:
            return self.checkpoint_forward(x)
        else:
            return self.fwd(x)


class RealFFT2(nn.Module):
    """
    Helper routine to wrap FFT similarly to the SHT
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None):
        super(RealFFT2, self).__init__()

        # use local FFT here
        self.fft_handle = torch.fft.rfft2

        self.nlat = nlat
        self.nlon = nlon
        self.lmax = min(lmax or self.nlat, self.nlat)
        self.mmax = min(mmax or self.nlon // 2 + 1, self.nlon // 2 + 1)

        self.truncate = True
        if (self.lmax == self.nlat) and (self.mmax == (self.nlon // 2 + 1)):
            self.truncate = False

        self.lmax_high = math.ceil(self.lmax / 2)
        self.lmax_low = math.floor(self.lmax / 2)

    def forward(self, x):
        y = self.fft_handle(x, s=(self.nlat, self.nlon), dim=(-2, -1), norm="ortho")

        if self.truncate:
            y = torch.cat(
                (
                    y[..., : self.lmax_high, : self.mmax],
                    y[..., -self.lmax_low :, : self.mmax],
                ),
                dim=-2,
            )

        return y


class InverseRealFFT2(nn.Module):
    """
    Helper routine to wrap FFT similarly to the SHT
    """

    def __init__(self, nlat, nlon, lmax=None, mmax=None):
        super(InverseRealFFT2, self).__init__()

        # use local FFT here
        self.ifft_handle = torch.fft.irfft2

        self.nlat = nlat
        self.nlon = nlon
        self.lmax = min(lmax or self.nlat, self.nlat)
        self.mmax = min(mmax or self.nlon // 2 + 1, self.nlon // 2 + 1)

        self.truncate = True
        if (self.lmax == self.nlat) and (self.mmax == (self.nlon // 2 + 1)):
            self.truncate = False

        self.lmax_high = math.ceil(self.lmax / 2)
        self.lmax_low = math.floor(self.lmax / 2)

    def forward(self, x):
        # truncation is implicit but better do it manually
        xt = x[..., : self.mmax]

        if self.truncate:
            # pad
            xth = xt[..., : self.lmax_high, :]
            xtl = xt[..., -self.lmax_low :, :]
            xthp = F.pad(xth, (0, 0, 0, self.nlat - self.lmax))
            xt = torch.cat([xthp, xtl], dim=-2)

        out = torch.fft.irfft2(xt, s=(self.nlat, self.nlon), dim=(-2, -1), norm="ortho")

        return out
